from concurrent.futures import Future
from threading import Event

import pytest

import parsl
from parsl.config import Config
from parsl.dataflow.dflow import DataFlowKernel, DataFlowKernelLoader
from parsl.errors import NoDataFlowKernelError
from parsl.tests.configs.local_threads import fresh_config


@parsl.python_app
def square(x):
    return x * x


@parsl.bash_app
def foo(x, stdout='foo.stdout'):
    return f"echo {x + 1}"


@parsl.python_app
def wait_for_event(ev: Event):
    ev.wait()


@parsl.python_app
def raise_app():
    raise RuntimeError("raise_app deliberate failure")


@pytest.mark.local
def test_within_context_manger(tmpd_cwd):
    config = fresh_config()
    with parsl.load(config=config) as dfk:
        assert isinstance(dfk, DataFlowKernel)

        bash_future = foo(1, stdout=tmpd_cwd / 'foo.stdout')
        assert bash_future.result() == 0

        with open(tmpd_cwd / 'foo.stdout', 'r') as f:
            assert f.read() == "2\n"

    with pytest.raises(NoDataFlowKernelError) as excinfo:
        square(2).result()
    assert str(excinfo.value) == "Must first load config"


@pytest.mark.local
def test_exit_skip():
    config = fresh_config()
    config.exit_mode = "skip"

    with parsl.load(config) as dfk:
        ev = Event()
        fut = wait_for_event(ev)
        # deliberately don't wait for this to finish, so that the context
        # manager can exit

    assert parsl.dfk() is dfk, "global dfk should be left in place by skip mode"

    assert not fut.done(), "wait_for_event should not be done yet"
    ev.set()

    # now we can wait for that result...
    fut.result()
    assert fut.done(), "wait_for_event should complete outside of context manager in 'skip' mode"

    # now cleanup the DFK that the above `with` block
    # deliberately avoided doing...
    dfk.cleanup()


# 'wait' mode has two cases to test:
# 1. that we wait when there is no exception
# 2. that we do not wait when there is an exception
@pytest.mark.local
def test_exit_wait_no_exception():
    config = fresh_config()
    config.exit_mode = "wait"

    with parsl.load(config) as dfk:
        fut = square(1)
        # deliberately don't wait for this to finish, so that the context
        # manager can exit

    assert fut.done(), "This future should be marked as done before the context manager exits"

    assert dfk.cleanup_called, "The DFK should have been cleaned up by the context manager"
    assert DataFlowKernelLoader._dfk is None, "The global DFK should have been removed"


@pytest.mark.local
def test_exit_wait_exception():
    config = fresh_config()
    config.exit_mode = "wait"

    with pytest.raises(RuntimeError):
        with parsl.load(config) as dfk:
            # we'll never fire this future
            fut_never = Future()

            fut_raise = raise_app()

            fut_depend = square(fut_never)

            # this should cause an exception, which should cause the context
            # manager to exit, without waiting for fut_depend to finish.
            fut_raise.result()

    assert dfk.cleanup_called, "The DFK should have been cleaned up by the context manager"
    assert DataFlowKernelLoader._dfk is None, "The global DFK should have been removed"
    assert fut_raise.exception() is not None, "fut_raise should contain an exception"
    assert not fut_depend.done(), "fut_depend should have been left un-done (due to dependency failure)"


@pytest.mark.local
def test_exit_wrong_mode():

    with pytest.raises(Exception) as ex:
        Config(exit_mode="wrongmode")

    # with typeguard 4.x this is TypeCheckError,
    # with typeguard 2.x this is TypeError
    # we can't instantiate TypeCheckError if we're in typeguard 2.x environment
    # because it does not exist... so check name using strings.
    assert ex.type.__name__ == "TypeCheckError" or ex.type.__name__ == "TypeError"



import pytest

from parsl.app.app import python_app
from parsl.dataflow.errors import DependencyError


class ManufacturedTestFailure(Exception):
    pass


@python_app
def random_fail(fail_prob: float, inputs=()):
    import random
    if random.random() < fail_prob:
        raise ManufacturedTestFailure("App failure")


def test_no_deps():
    """Test basic error handling, with no dependent failures
    """
    futs = [random_fail(1), random_fail(0), random_fail(0)]

    for f in futs:
        try:
            f.result()
        except ManufacturedTestFailure:
            pass


def test_fail_sequence_first():
    t1 = random_fail(fail_prob=1)
    t2 = random_fail(fail_prob=0, inputs=[t1])
    t_final = random_fail(fail_prob=0, inputs=[t2])

    with pytest.raises(DependencyError):
        t_final.result()

    assert len(t_final.exception().dependent_exceptions_tids) == 1
    assert isinstance(t_final.exception().dependent_exceptions_tids[0][0], DependencyError)
    assert t_final.exception().dependent_exceptions_tids[0][1].startswith("task ")

    assert hasattr(t_final.exception(), '__cause__')
    assert t_final.exception().__cause__ == t1.exception()


def test_fail_sequence_middle():
    t1 = random_fail(fail_prob=0)
    t2 = random_fail(fail_prob=1, inputs=[t1])
    t_final = random_fail(fail_prob=0, inputs=[t2])

    with pytest.raises(DependencyError):
        t_final.result()

    assert len(t_final.exception().dependent_exceptions_tids) == 1
    assert isinstance(t_final.exception().dependent_exceptions_tids[0][0], ManufacturedTestFailure)

    assert hasattr(t_final.exception(), '__cause__')
    assert t_final.exception().__cause__ == t2.exception()



import pytest

from parsl.app.app import python_app
from parsl.data_provider.files import File
from parsl.tests.configs.local_threads_ftp_in_task import fresh_config as local_config


@python_app
def sort_strings(inputs=[], outputs=[]):
    with open(inputs[0].filepath, 'r') as u:
        strs = u.readlines()
        strs.sort()
        with open(outputs[0].filepath, 'w') as s:
            for e in strs:
                s.write(e)


@pytest.mark.local
@pytest.mark.cleannet
def test_staging_ftp():
    """Test staging for an ftp file

    Create a remote input file (ftp) that points to file_test_cpt.txt.
    """

    unsorted_file = File('ftp://ftp.iana.org/pub/mirror/rirstats/arin/ARIN-STATS-FORMAT-CHANGE.txt')

    # Create a local file for output data
    sorted_file = File('sorted.txt')

    f = sort_strings(inputs=[unsorted_file], outputs=[sorted_file])
    f.result()



import pytest

from parsl.app.app import bash_app, python_app
from parsl.data_provider.files import File


@bash_app
def generate(outputs=()):
    return "echo 1 &> {o}".format(o=outputs[0])


@bash_app
def concat(inputs=(), outputs=(), stdout=None, stderr=None):
    return "cat {0} >> {1}".format(" ".join(map(lambda x: x.filepath, inputs)), outputs[0])


@python_app
def total(inputs=()):
    with open(inputs[0].filepath, "r") as f:
        return sum(int(line) for line in f)


@pytest.mark.staging_required
@pytest.mark.parametrize("width", (5, 10, 15))
def test_parallel_dataflow(tmpd_cwd, width):
    """Test parallel dataflow from docs on Composing workflows
    """

    # create 5 files with random numbers
    output_files = [
        generate(outputs=[File(str(tmpd_cwd / f"random-{i}.txt"))])
        for i in range(width)
    ]

    # concatenate the files into a single file
    cc = concat(
        inputs=[i.outputs[0] for i in output_files],
        outputs=[File(str(tmpd_cwd / "all.txt"))]
    )

    # calculate the average of the random numbers
    totals = total(inputs=[cc.outputs[0]])
    assert totals.result() == len(output_files)



import logging
import uuid

import pytest

import parsl
from parsl.serialize.facade import (
    additional_methods_for_deserialization,
    methods_for_data,
    register_method_for_data,
)
from parsl.tests.configs.htex_local import fresh_config

logger = logging.getLogger(__name__)


def local_setup():
    from proxystore.connectors.file import FileConnector
    from proxystore.store import Store, register_store

    from parsl.serialize.proxystore import ProxyStoreSerializer

    parsl.load(fresh_config())

    store = Store(name='parsl_store_' + str(uuid.uuid4()), connector=FileConnector(store_dir="/tmp"))
    register_store(store)

    s = ProxyStoreSerializer(store=store, should_proxy=policy_example)

    global previous_methods
    previous_methods = methods_for_data.copy()

    # get rid of all data serialization methods, in preparation for using only
    # proxystore. put all the old methods as additional methods used only for
    # deserialization, because those will be needed to deserialize the results,
    # which will be serialized using the default serializer set.
    additional_methods_for_deserialization.update(previous_methods)
    methods_for_data.clear()

    register_method_for_data(s)
    logger.info(f"BENC: methods for data: {methods_for_data}")


def local_teardown():
    parsl.dfk().cleanup()

    methods_for_data.clear()
    methods_for_data.update(previous_methods)

    additional_methods_for_deserialization.clear()


@parsl.python_app
def identity(o):
    return o


@parsl.python_app
def is_proxy(o) -> bool:
    from proxystore.proxy import Proxy
    return isinstance(o, Proxy)


def policy_example(o):
    """Example policy will proxy only lists."""
    return isinstance(o, frozenset)


@pytest.mark.local
def test_proxystore_via_apps():
    from proxystore.proxy import Proxy

    # check roundtrip for an int, which should not be proxystored according
    # to example_policy()
    roundtripped_7 = identity(7).result()
    assert roundtripped_7 == 7
    assert not isinstance(roundtripped_7, Proxy)

    # check roundtrip for a list, which should be proxystored according
    # to example_policy()
    v = frozenset([1, 2, 3])

    assert is_proxy(v).result()

    roundtripped_v = identity(v).result()
    assert roundtripped_v == v



"""Tests for the wrapper class"""
from inspect import signature
from pathlib import Path

import pytest

from parsl import Config, HighThroughputExecutor
from parsl.executors.high_throughput.mpi_executor import MPIExecutor
from parsl.launchers import SimpleLauncher
from parsl.providers import LocalProvider

from .test_mpi_mode_enabled import get_env_vars

cwd = Path(__file__).parent.absolute()
pbs_nodefile = cwd.joinpath("mocks", "pbs_nodefile")


def local_config():
    return Config(
        executors=[
            MPIExecutor(
                max_workers_per_block=1,
                provider=LocalProvider(
                    worker_init=f"export PBS_NODEFILE={pbs_nodefile}",
                    launcher=SimpleLauncher()
                )
            )
        ]
    )


@pytest.mark.local
def test_docstring():
    """Ensure the old kwargs are copied over into the new class"""
    assert 'label' in MPIExecutor.__doc__
    assert 'max_workers_per_block' in MPIExecutor.__doc__
    assert 'available_accelerators' not in MPIExecutor.__doc__


@pytest.mark.local
def test_init():
    """Ensure all relevant kwargs are copied over from HTEx"""

    new_kwargs = {'max_workers_per_block', 'mpi_launcher'}
    excluded_kwargs = {'available_accelerators', 'cores_per_worker', 'max_workers_per_node',
                       'mem_per_worker', 'cpu_affinity', 'manager_selector'}

    # Get the kwargs from both HTEx and MPIEx
    htex_kwargs = set(signature(HighThroughputExecutor.__init__).parameters)
    mpix_kwargs = set(signature(MPIExecutor.__init__).parameters)

    assert mpix_kwargs.difference(htex_kwargs) == new_kwargs
    assert len(mpix_kwargs.intersection(excluded_kwargs)) == 0
    assert mpix_kwargs.union(excluded_kwargs).difference(new_kwargs) == htex_kwargs


@pytest.mark.local
def test_get_env():
    future = get_env_vars(parsl_resource_specification={
        "num_nodes": 3,
        "ranks_per_node": 5,
    })
    env_vars = future.result()
    assert env_vars['PARSL_NUM_RANKS'] == '15'



from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.providers import LocalProvider

# import os
config = Config(
    executors=[
        HighThroughputExecutor(
            label="htex_local",
            cores_per_worker=1,
            max_workers_per_node=8,
            encrypted=True,
            provider=LocalProvider(
                init_blocks=1,
                max_blocks=1,
            ),
        )
    ],
    strategy='none',
)



import logging
import multiprocessing
import os
import pickle
import queue
import subprocess
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Optional

from parsl.multiprocessing import SpawnContext
from parsl.serialize import pack_res_spec_apply_message, unpack_res_spec_apply_message

logger = logging.getLogger(__name__)


class Scheduler(Enum):
    Unknown = 0
    Slurm = 1
    PBS = 2


def get_slurm_hosts_list() -> List[str]:
    """Get list of slurm hosts from scontrol"""
    cmd = "scontrol show hostname $SLURM_NODELIST"
    b_output = subprocess.check_output(
        cmd, stderr=subprocess.STDOUT, shell=True
    )  # bytes
    output = b_output.decode().strip().split()
    return output


def get_pbs_hosts_list() -> List[str]:
    """Get list of PBS hosts from envvar: PBS_NODEFILE"""
    nodefile_name = os.environ["PBS_NODEFILE"]
    with open(nodefile_name) as f:
        return [line.strip() for line in f.readlines()]


def get_nodes_in_batchjob(scheduler: Scheduler) -> List[str]:
    """Get nodelist from all supported schedulers"""
    nodelist = []
    if scheduler == Scheduler.Slurm:
        nodelist = get_slurm_hosts_list()
    elif scheduler == Scheduler.PBS:
        nodelist = get_pbs_hosts_list()
    else:
        raise RuntimeError(f"mpi_mode does not support scheduler:{scheduler}")
    return nodelist


def identify_scheduler() -> Scheduler:
    """Use envvars to determine batch scheduler"""
    if os.environ.get("SLURM_NODELIST"):
        return Scheduler.Slurm
    elif os.environ.get("PBS_NODEFILE"):
        return Scheduler.PBS
    else:
        return Scheduler.Unknown


class MPINodesUnavailable(Exception):
    """Raised if there are no free nodes available for an MPI request"""

    def __init__(self, requested: int, available: int):
        self.requested = requested
        self.available = available

    def __str__(self):
        return f"MPINodesUnavailable(requested={self.requested} available={self.available})"


@dataclass(order=True)
class PrioritizedTask:
    # Comparing dict will fail since they are unhashable
    # This dataclass limits comparison to the priority field
    priority: int
    task: Dict = field(compare=False)


class TaskScheduler:
    """Default TaskScheduler that does no taskscheduling

    This class simply acts as an abstraction over the task_q and result_q
    that can be extended to implement more complex task scheduling logic
    """
    def __init__(
        self,
        pending_task_q: multiprocessing.Queue,
        pending_result_q: multiprocessing.Queue,
    ):
        self.pending_task_q = pending_task_q
        self.pending_result_q = pending_result_q

    def put_task(self, task) -> None:
        return self.pending_task_q.put(task)

    def get_result(self, block: bool = True, timeout: Optional[float] = None):
        return self.pending_result_q.get(block, timeout)


class MPITaskScheduler(TaskScheduler):
    """Extends TaskScheduler to schedule MPI functions over provisioned nodes
    The MPITaskScheduler runs on a Manager on the lead node of a batch job, as
    such it is expected to control task placement over this single batch job.

    The MPITaskScheduler adds the following functionality:
    1) Determine list of nodes attached to current batch job
    2) put_task for execution onto workers:
        a) if resources are available attach resource list
        b) if unavailable place tasks into backlog
    3) get_result will fetch a result and relinquish nodes,
       and attempt to schedule tasks in backlog if any.
    """
    def __init__(
        self,
        pending_task_q: multiprocessing.Queue,
        pending_result_q: multiprocessing.Queue,
    ):
        super().__init__(pending_task_q, pending_result_q)
        self.scheduler = identify_scheduler()
        # PriorityQueue is threadsafe
        self._backlog_queue: queue.PriorityQueue[PrioritizedTask] = queue.PriorityQueue()
        self._map_tasks_to_nodes: Dict[str, List[str]] = {}
        self.available_nodes = get_nodes_in_batchjob(self.scheduler)
        self._free_node_counter = SpawnContext.Value("i", len(self.available_nodes))
        # mp.Value has issues with mypy
        # issue https://github.com/python/typeshed/issues/8799
        # from mypy 0.981 onwards
        self.nodes_q: queue.Queue = queue.Queue()
        for node in self.available_nodes:
            self.nodes_q.put(node)

        logger.info(
            f"Starting MPITaskScheduler with {len(self.available_nodes)}"
        )

    def _get_nodes(self, num_nodes: int) -> List[str]:
        """Thread safe method to acquire num_nodes from free resources

        Raises: MPINodesUnavailable if there aren't enough resources
        Returns: List of nodenames:str
        """
        logger.debug(
            f"Requesting : {num_nodes=} we have {self._free_node_counter}"
        )
        acquired_nodes = []
        with self._free_node_counter.get_lock():
            if num_nodes <= self._free_node_counter.value:  # type: ignore[attr-defined]
                self._free_node_counter.value -= num_nodes  # type: ignore[attr-defined]
            else:
                raise MPINodesUnavailable(
                    requested=num_nodes, available=self._free_node_counter.value  # type: ignore[attr-defined]
                )

            for i in range(num_nodes):
                node = self.nodes_q.get()
                acquired_nodes.append(node)
        return acquired_nodes

    def _return_nodes(self, nodes: List[str]) -> None:
        """Threadsafe method to return a list of nodes"""
        for node in nodes:
            self.nodes_q.put(node)
        with self._free_node_counter.get_lock():
            self._free_node_counter.value += len(nodes)  # type: ignore[attr-defined]

    def put_task(self, task_package: dict):
        """Schedule task if resources are available otherwise backlog the task"""
        user_ns = locals()
        user_ns.update({"__builtins__": __builtins__})
        _f, _args, _kwargs, resource_spec = unpack_res_spec_apply_message(task_package["buffer"])

        nodes_needed = resource_spec.get("num_nodes")
        tid = task_package["task_id"]
        if nodes_needed:
            try:
                allocated_nodes = self._get_nodes(nodes_needed)
            except MPINodesUnavailable:
                logger.info(f"Not enough resources, placing task {tid} into backlog")
                self._backlog_queue.put(PrioritizedTask(nodes_needed, task_package))
                return
            else:
                resource_spec["MPI_NODELIST"] = ",".join(allocated_nodes)
                self._map_tasks_to_nodes[tid] = allocated_nodes
                buffer = pack_res_spec_apply_message(_f, _args, _kwargs, resource_spec)
                task_package["buffer"] = buffer
                task_package["resource_spec"] = resource_spec

        self.pending_task_q.put(task_package)

    def _schedule_backlog_tasks(self):
        """Attempt to schedule backlogged tasks"""
        try:
            prioritized_task = self._backlog_queue.get(block=False)
            self.put_task(prioritized_task.task)
        except queue.Empty:
            return
        else:
            # Keep attempting to schedule tasks till we are out of resources
            self._schedule_backlog_tasks()

    def get_result(self, block: bool = True, timeout: Optional[float] = None):
        """Return result and relinquish provisioned nodes"""
        result_pkl = self.pending_result_q.get(block, timeout)
        if result_pkl is None:
            return None
        result_dict = pickle.loads(result_pkl)
        # TODO (wardlt): If the task did not request nodes, it won't be in `self._map_tasks_to_nodes`.
        #  Causes Parsl to hang. See Issue #3427
        if result_dict["type"] == "result":
            task_id = result_dict["task_id"]
            assert task_id in self._map_tasks_to_nodes, "You are about to experience issue #3427"
            nodes_to_reallocate = self._map_tasks_to_nodes[task_id]
            self._return_nodes(nodes_to_reallocate)
            self._schedule_backlog_tasks()

        return result_pkl



"""Test usage_tracking values."""

import pytest

import parsl
from parsl.config import Config
from parsl.errors import ConfigurationError


@pytest.mark.local
def test_config_load():
    """Test loading a config with usage tracking."""
    with parsl.load(Config(usage_tracking=3)):
        pass
    parsl.clear()


@pytest.mark.local
@pytest.mark.parametrize("level", (0, 1, 2, 3, False, True))
def test_valid(level):
    """Test valid usage_tracking values."""
    Config(usage_tracking=level)
    assert Config(usage_tracking=level).usage_tracking == level


@pytest.mark.local
@pytest.mark.parametrize("level", (12, 1000, -1))
def test_invalid_values(level):
    """Test invalid usage_tracking values."""
    with pytest.raises(ConfigurationError):
        Config(usage_tracking=level)


@pytest.mark.local
@pytest.mark.parametrize("level", ("abcd", None, bytes(1), 1.0, 1j, object()))
def test_invalid_types(level):
    """Test invalid usage_tracking types."""
    with pytest.raises(Exception) as ex:
        Config(usage_tracking=level)

    # with typeguard 4.x this is TypeCheckError,
    # with typeguard 2.x this is TypeError
    # we can't instantiate TypeCheckError if we're in typeguard 2.x environment
    # because it does not exist... so check name using strings.
    assert ex.type.__name__ in ["TypeCheckError", "TypeError"]


@pytest.mark.local
def test_valid_project_name():
    """Test valid project_name."""
    assert (
        Config(
            usage_tracking=3,
            project_name="unit-test",
        ).project_name == "unit-test"
    )


@pytest.mark.local
@pytest.mark.parametrize("name", (1, 1.0, True, object()))
def test_invalid_project_name(name):
    """Test invalid project_name."""
    with pytest.raises(Exception) as ex:
        Config(usage_tracking=3, project_name=name)

    assert ex.type.__name__ in ["TypeCheckError", "TypeError"]



""" TaskVineExecutor utilizes the TaskVine distributed framework developed by the
Cooperative Computing Lab (CCL) at Notre Dame to provide a fault-tolerant,
high-throughput system for delegating Parsl tasks to thousands of remote machines
"""

import getpass
import hashlib
import inspect
import itertools
import logging
import multiprocessing
import os
import queue
import shutil
import subprocess
import tempfile

# Import Python built-in libraries
import threading
import uuid
from concurrent.futures import Future
from datetime import datetime
from typing import List, Literal, Optional, Union

# Import other libraries
import typeguard

# Import Parsl constructs
import parsl.utils as putils
from parsl.addresses import get_any_address
from parsl.data_provider.files import File
from parsl.data_provider.staging import Staging
from parsl.errors import OptionalModuleMissing
from parsl.executors.errors import ExecutorError
from parsl.executors.status_handling import BlockProviderExecutor
from parsl.executors.taskvine import exec_parsl_function
from parsl.executors.taskvine.errors import TaskVineManagerFailure, TaskVineTaskFailure
from parsl.executors.taskvine.factory import _taskvine_factory
from parsl.executors.taskvine.factory_config import TaskVineFactoryConfig
from parsl.executors.taskvine.manager import _taskvine_submit_wait
from parsl.executors.taskvine.manager_config import TaskVineManagerConfig
from parsl.executors.taskvine.utils import ParslFileToVine, ParslTaskToVine
from parsl.monitoring.radios.base import RadioConfig
from parsl.monitoring.radios.filesystem import FilesystemRadio
from parsl.multiprocessing import SpawnContext
from parsl.process_loggers import wrap_with_logs
from parsl.providers import CondorProvider, LocalProvider
from parsl.providers.base import ExecutionProvider
from parsl.serialize import deserialize, serialize

logger = logging.getLogger(__name__)


class TaskVineExecutor(BlockProviderExecutor, putils.RepresentationMixin):
    """Executor to use TaskVine dynamic workflow system

    The TaskVineExecutor system utilizes the TaskVine framework to
    efficiently delegate Parsl apps to remote machines in clusters and
    grids using a fault-tolerant system. Users can run the
    vine_worker program on remote machines to connect to the
    TaskVineExecutor, and Parsl apps will then be sent out to these
    machines for execution and retrieval.

    This Executor sets up configurations for the TaskVine manager, TaskVine
    factory, and run both in separate processes. Sending tasks and receiving
    results are done through multiprocessing module native to Python.

    Parameters
    ----------

        label: str
            A human readable label for the executor, unique
            with respect to other executors.
            Default is "TaskVineExecutor".

        worker_launch_method: Union[Literal['provider'], Literal['factory'], Literal['manual']]
            Choose to use Parsl provider, TaskVine factory, or
            manual user-provided workers to scale workers.
            Options are among {'provider', 'factory', 'manual'}.
            Default is 'factory'.

        function_exec_mode: Union[Literal['regular'], Literal['serverless']]
            Choose to execute functions with a regular fresh python process or a
            pre-warmed forked python process.
            Default is 'regular'.

        manager_config: TaskVineManagerConfig
            Configuration for the TaskVine manager. Default

        factory_config: TaskVineFactoryConfig
            Configuration for the TaskVine factory.
            Use of factory is disabled by default.

        provider: ExecutionProvider
            The Parsl provider that will spawn worker processes.
            Default to spawning one local vine worker process.

        storage_access: List[Staging]
            Define Parsl file staging providers for this executor.
            Default is None.
    """

    @typeguard.typechecked
    def __init__(self,
                 label: str = "TaskVineExecutor",
                 worker_launch_method: Union[Literal['provider'], Literal['factory'], Literal['manual']] = 'factory',
                 function_exec_mode: Union[Literal['regular'], Literal['serverless']] = 'regular',
                 manager_config: TaskVineManagerConfig = TaskVineManagerConfig(),
                 factory_config: TaskVineFactoryConfig = TaskVineFactoryConfig(),
                 provider: Optional[ExecutionProvider] = LocalProvider(init_blocks=1),
                 storage_access: Optional[List[Staging]] = None,
                 remote_monitoring_radio: Optional[RadioConfig] = None):

        # Set worker launch option for this executor
        if worker_launch_method == 'factory' or worker_launch_method == 'manual':
            provider = None

        # Initialize the parent class with the execution provider and block error handling enabled.
        # If provider is None, then no worker is launched via the provider method.
        BlockProviderExecutor.__init__(self, provider=provider,
                                       block_error_handler=True)

        # Raise an exception if there's a problem importing TaskVine
        try:
            import ndcctools.taskvine
            logger.debug(f'TaskVine default port: {ndcctools.taskvine.cvine.VINE_DEFAULT_PORT}')
        except ImportError:
            raise OptionalModuleMissing(['taskvine'], "TaskVineExecutor requires the taskvine module.")

        # Executor configurations
        self.label = label
        self.worker_launch_method = worker_launch_method
        self.function_exec_mode = function_exec_mode
        self.manager_config = manager_config
        self.factory_config = factory_config
        self.storage_access = storage_access

        if remote_monitoring_radio is not None:
            self.remote_monitoring_radio = remote_monitoring_radio
        else:
            self.remote_monitoring_radio = FilesystemRadio()

        # Queue to send ready tasks from TaskVine executor process to TaskVine manager process
        self._ready_task_queue: multiprocessing.Queue = SpawnContext.Queue()

        # Queue to send finished tasks from TaskVine manager process to TaskVine executor process
        self._finished_task_queue: multiprocessing.Queue = SpawnContext.Queue()

        # Event to signal whether the manager and factory processes should stop running
        self._should_stop = SpawnContext.Event()

        # TaskVine manager process
        self._submit_process = None

        # TaskVine factory process
        self._factory_process = None

        # Executor thread to collect results from TaskVine manager and set
        # tasks' futures to done status.
        self._collector_thread = None

        # track task id of submitted parsl tasks
        # task ids are incremental and start from 0
        self._executor_task_counter = 0

        # track number of tasks that are waiting/running
        self._outstanding_tasks = 0

        # Lock for threads to access self._outstanding_tasks attribute
        self._outstanding_tasks_lock = threading.Lock()

        # Threading lock to manage self.tasks dictionary object, which maps a task id
        # to its future object.
        self._tasks_lock = threading.Lock()

        # Worker command to be given to an execution provider (e.g., local or Condor)
        self._worker_command = ""

        # Path to directory that holds all tasks' data and results.
        self._function_data_dir = ""

        # Helper scripts to prepare package tarballs for Parsl apps
        self._package_analyze_script = shutil.which("poncho_package_analyze")
        self._package_create_script = shutil.which("poncho_package_create")
        if self._package_analyze_script is None or self._package_create_script is None:
            self._poncho_available = False
        else:
            self._poncho_available = True

    def _get_launch_command(self, block_id):
        # Implements BlockProviderExecutor's abstract method.
        # This executor uses different terminology for worker/launch
        # commands than in htex.
        return f"PARSL_WORKER_BLOCK_ID={block_id} {self._worker_command}"

    def __synchronize_manager_factory_comm_settings(self):
        # Synchronize the communication settings between the manager and the factory
        # so the factory can direct workers to contact the manager.

        # If the manager can choose any available port (port number = 0),
        # then it must have a project name
        # so the factory can look it up. Otherwise the factory process will not know the
        # port number as it's only chosen when the TaskVine manager process is run.
        if self.manager_config.port == 0 and self.manager_config.project_name is None:
            self.manager_config.project_name = "parsl-vine-" + str(uuid.uuid4())

        # guess the host name if the project name is not given and none has been supplied
        # explicitly in the manager config.
        if not self.manager_config.project_name and self.manager_config.address is None:
            self.manager_config.address = get_any_address()

        # Factory communication settings are overridden by manager communication settings.
        self.factory_config._project_port = self.manager_config.port
        self.factory_config._project_address = self.manager_config.address
        self.factory_config._project_name = self.manager_config.project_name
        self.factory_config._project_password_file = self.manager_config.project_password_file
        logger.debug('Communication settings between TaskVine manager and factory synchronized.')

    def __create_data_and_logging_dirs(self):
        # Create neccessary data and logging directories

        # Use the current run directory from Parsl
        run_dir = self.run_dir

        # Create directories for data and results
        log_dir = os.path.join(run_dir, self.label)
        os.makedirs(log_dir)
        tmp_prefix = f'{self.label}-{getpass.getuser()}-{datetime.now().strftime("%Y%m%d%H%M%S%f")}-'
        self._function_data_dir = tempfile.TemporaryDirectory(prefix=tmp_prefix)

        # put TaskVine logs outside of a Parsl run as TaskVine caches between runs while
        # Parsl does not.
        vine_log_dir = os.path.join(os.path.dirname(run_dir), self.label)
        if self.manager_config.vine_log_dir is None:
            self.manager_config.vine_log_dir = vine_log_dir

        # factory logs go with manager logs regardless
        self.factory_config.scratch_dir = self.manager_config.vine_log_dir
        logger.debug(f"Function data directory: {self._function_data_dir.name}, log directory: {log_dir}")
        logger.debug(
            f"TaskVine manager log directory: {self.manager_config.vine_log_dir}, "
            f"factory log directory: {self.factory_config.scratch_dir}")

    def start(self):
        """Create submit process and collector thread to create, send, and
        retrieve Parsl tasks within the TaskVine system.
        """

        super().start()
        # Synchronize connection and communication settings between the manager and factory
        self.__synchronize_manager_factory_comm_settings()

        # Create data and logging dirs
        self.__create_data_and_logging_dirs()

        logger.debug("Starting TaskVineExecutor")

        # Create a process to run the TaskVine manager.
        submit_process_kwargs = {"ready_task_queue": self._ready_task_queue,
                                 "finished_task_queue": self._finished_task_queue,
                                 "should_stop": self._should_stop,
                                 "manager_config": self.manager_config}
        self._submit_process = SpawnContext.Process(target=_taskvine_submit_wait,
                                                    name="TaskVine-Submit-Process",
                                                    kwargs=submit_process_kwargs)

        # Create a process to run the TaskVine factory if enabled.
        if self.worker_launch_method == 'factory':
            factory_process_kwargs = {"should_stop": self._should_stop,
                                      "factory_config": self.factory_config}
            self._factory_process = SpawnContext.Process(target=_taskvine_factory,
                                                         name="TaskVine-Factory-Process",
                                                         kwargs=factory_process_kwargs)

        # Run thread to collect results and set tasks' futures.
        self._collector_thread = threading.Thread(target=self._collect_taskvine_results,
                                                  name="TaskVine-Collector-Thread")
        # Interpreter can exit without waiting for this thread.
        self._collector_thread.daemon = True

        # Begin work
        self._submit_process.start()

        # Run worker scaler either with Parsl provider or TaskVine factory.
        # Skip if workers are launched manually.
        if self.worker_launch_method == 'factory':
            self._factory_process.start()
        elif self.worker_launch_method == 'provider':
            self.initialize_scaling()

        self._collector_thread.start()

        logger.debug("All components in TaskVineExecutor started")

    def _path_in_task(self, executor_task_id, *path_components):
        """
        Returns a filename fixed and specific to a task.
        It is used for the following filename's:
            (not given): The subdirectory per task that contains function, result, etc.
            'function': Pickled file that contains the function to be executed.
            'argument': Pickled file that contains the arguments of the function call.
            'result': Pickled file that (will) contain the result of the function.
            'map': Pickled file with a dict between local parsl names, and remote taskvine names.
        """
        task_dir = "{:04d}".format(executor_task_id)
        return os.path.join(self._function_data_dir.name, task_dir, *path_components)

    def submit(self, func, resource_specification, *args, **kwargs):
        """Processes the Parsl app by its arguments and submits the function
        information to the task queue, to be executed using the TaskVine
        system. The args and kwargs are processed for input and output files to
        the Parsl app, so that the files are appropriately specified for the TaskVine task.

        Parameters
        ----------

        func : function
            Parsl app to be submitted to the TaskVine system
        resource_specification: dict
            Dictionary containing relevant info about task.
            Include information about resources of task, execution mode
            of task (out of {regular, serverless}).
        args : list
            Arguments to the Parsl app
        kwargs : dict
            Keyword arguments to the Parsl app
        """

        logger.debug(f'Got resource specification: {resource_specification}')

        # Default execution mode of apps is regular
        exec_mode = resource_specification.get('exec_mode', self.function_exec_mode)

        # Detect resources and features of a submitted Parsl app
        cores = None
        memory = None
        disk = None
        gpus = None
        priority = None
        category = None
        running_time_min = None
        if resource_specification and isinstance(resource_specification, dict):
            for k in resource_specification:
                if k == 'cores':
                    cores = resource_specification[k]
                elif k == 'memory':
                    memory = resource_specification[k]
                elif k == 'disk':
                    disk = resource_specification[k]
                elif k == 'gpus':
                    gpus = resource_specification[k]
                elif k == 'priority':
                    priority = resource_specification[k]
                elif k == 'category':
                    category = resource_specification[k]
                elif k == 'running_time_min':
                    running_time_min = resource_specification[k]

        # Assign executor task id to app
        executor_task_id = self._executor_task_counter
        self._executor_task_counter += 1

        # Create a per task directory for the function, argument, map, and result files
        os.mkdir(self._path_in_task(executor_task_id))

        input_files = []
        output_files = []

        # Determine whether to stage input files that will exist at the workers
        # Input and output files are always cached
        input_files += [self._register_file(f) for f in kwargs.get("inputs", []) if isinstance(f, File)]
        output_files += [self._register_file(f) for f in kwargs.get("outputs", []) if isinstance(f, File)]

        # Also consider any *arg that looks like a file as an input:
        input_files += [self._register_file(f) for f in args if isinstance(f, File)]

        for kwarg, maybe_file in kwargs.items():
            # Add appropriate input and output files from "stdout" and "stderr" keyword arguments
            if kwarg == "stdout" or kwarg == "stderr":
                if maybe_file:
                    output_files.append(self._std_output_to_vine(kwarg, maybe_file))
            # For any other keyword that looks like a file, assume it is an input file
            elif isinstance(maybe_file, File):
                input_files.append(self._register_file(maybe_file))

        # Create a Future object and have it be mapped from the task ID in the tasks dictionary
        fu = Future()
        fu.parsl_executor_task_id = executor_task_id
        with self._tasks_lock:
            self.tasks[str(executor_task_id)] = fu

        # Setup files to be used on a worker to execute the function
        function_file = None
        argument_file = None
        result_file = None
        map_file = None

        # Get path to files that will contain the pickled function,
        # arguments, result, and map of input and output files
        function_file = self._path_in_task(executor_task_id, "function")
        argument_file = self._path_in_task(executor_task_id, "argument")
        result_file = self._path_in_task(executor_task_id, "result")
        map_file = self._path_in_task(executor_task_id, "map")

        logger.debug("Creating executor task {} with function at: {}, argument at: {}, \
                and result to be found at: {}".format(executor_task_id, function_file, argument_file, result_file))

        # Serialize function object and arguments, separately
        self._serialize_object_to_file(function_file, func)
        args_dict = {'args': args, 'kwargs': kwargs}
        self._serialize_object_to_file(argument_file, args_dict)

        # Construct the map file of local filenames at worker
        self._construct_map_file(map_file, input_files, output_files)

        # Register a tarball containing all package dependencies for this app if instructed
        if self.manager_config.app_pack:
            env_pkg = self._prepare_package(func, self.extra_pkgs)
        else:
            env_pkg = None

        # Create message to put into the message queue
        logger.debug("Placing task {} on message queue".format(executor_task_id))

        # Put apps into their categories based on function name if enabled
        if category is None:
            category = func.__name__ if self.manager_config.autocategory else 'parsl-default'

        task_info = ParslTaskToVine(executor_id=executor_task_id,
                                    exec_mode=exec_mode,
                                    category=category,
                                    input_files=input_files,
                                    output_files=output_files,
                                    map_file=map_file,
                                    function_file=function_file,
                                    argument_file=argument_file,
                                    result_file=result_file,
                                    cores=cores,
                                    memory=memory,
                                    disk=disk,
                                    gpus=gpus,
                                    priority=priority,
                                    running_time_min=running_time_min,
                                    env_pkg=env_pkg)

        # Send ready task to manager process
        if not self._submit_process.is_alive():
            raise ExecutorError(self, "taskvine Submit Process is not alive")

        self._ready_task_queue.put_nowait(task_info)

        # Increment outstanding task counter
        with self._outstanding_tasks_lock:
            self._outstanding_tasks += 1

        # Return the future for this function, will be set by collector thread when result
        # comes back from the TaskVine manager.
        return fu

    def _construct_worker_command(self):
        worker_command = self.factory_config.worker_executable
        if self.factory_config._project_password_file:
            worker_command += ' --password {}'.format(self.factory_config._project_password_file)
        if self.factory_config.worker_options:
            worker_command += ' {}'.format(self.factory_config.worker_options)
        if self.factory_config._project_name:
            worker_command += ' -M {}'.format(self.factory_config._project_name)
        else:
            worker_command += ' {} {}'.format(self.factory_config._project_address, self.factory_config._project_port)

        logger.debug("Using worker command: {}".format(worker_command))
        return worker_command

    def _patch_providers(self):
        # Add the worker and password file to files that the provider needs to stage.
        # (Currently only for the CondorProvider)
        if isinstance(self.provider, CondorProvider):
            path_to_worker = shutil.which('vine_worker')
            self._worker_command = './' + self._worker_command
            self.provider.transfer_input_files.append(path_to_worker)
            if self.project_password_file:
                self.provider.transfer_input_files.append(self.project_password_file)

    def _serialize_object_to_file(self, path, obj):
        """Takes any object and serializes it to the file path."""
        serialized_obj = serialize(obj, buffer_threshold=1024 * 1024)
        with open(path, 'wb') as f_out:
            written = 0
            while written < len(serialized_obj):
                written += f_out.write(serialized_obj[written:])

    def _construct_map_file(self, map_file, input_files, output_files):
        """ Map local filepath of parsl files to the filenames at the execution worker.
        If using a shared filesystem, the filepath is mapped to its absolute filename.
        Otherwise, to its original relative filename. In this later case, taskvine
        recreates any directory hierarchy needed."""
        file_translation_map = {}
        for spec in itertools.chain(input_files, output_files):
            local_name = spec.parsl_name
            if self.manager_config.shared_fs:
                remote_name = os.path.abspath(local_name)
            else:
                remote_name = local_name
            file_translation_map[local_name] = remote_name
        self._serialize_object_to_file(map_file, file_translation_map)

    def _register_file(self, parsl_file):
        """Generates a tuple (parsl_file.filepath, stage, cache) to give to
        taskvine. cache is always True.
        stage is True if the file has a relative path. (i.e., not
        a URL or an absolute path)"""
        to_cache = True
        to_stage = False
        if parsl_file.scheme == 'file' or \
           (parsl_file.local_path and os.path.exists(parsl_file.local_path)):
            to_stage = not os.path.isabs(parsl_file.filepath)

        return ParslFileToVine(parsl_file.filepath, to_stage, to_cache)

    def _std_output_to_vine(self, fdname, stdfspec):
        """Find the name of the file that will contain stdout or stderr and
        return a ParslFileToVine with it. These files are never cached"""
        fname, mode = putils.get_std_fname_mode(fdname, stdfspec)
        to_stage = not os.path.isabs(fname)
        return ParslFileToVine(fname, stage=to_stage, cache=False)

    def _prepare_package(self, fn, extra_pkgs):
        """ Look at source code of apps to figure out their package depedencies
        and output a tarball containing those to send along with tasks for execution."""

        if not self._poncho_available:
            raise ExecutorError(self, 'poncho package is not available to individually pack apps.')

        fn_id = id(fn)
        fn_name = fn.__name__
        if fn_id in self.cached_envs:
            logger.debug("Skipping analysis of %s, previously got %s", fn_name, self.cached_envs[fn_id])
            return self.cached_envs[fn_id]
        source_code = inspect.getsource(fn).encode()
        pkg_dir = os.path.join(tempfile.gettempdir(), "python_package-{}".format(os.geteuid()))
        os.makedirs(pkg_dir, exist_ok=True)
        with tempfile.NamedTemporaryFile(suffix='.yaml') as spec:
            logger.info("Analyzing dependencies of %s", fn_name)
            analyze_cmdline = [self._package_analyze_script, exec_parsl_function.__file__, '-', spec.name]
            for p in extra_pkgs:
                analyze_cmdline += ["--extra-pkg", p]
            subprocess.run(analyze_cmdline, input=source_code, check=True)
            with open(spec.name, mode='rb') as f:
                spec_hash = hashlib.sha256(f.read()).hexdigest()
                logger.debug("Spec hash for %s is %s", fn_name, spec_hash)
                pkg = os.path.join(pkg_dir, "pack-{}.tar.gz".format(spec_hash))
            if os.access(pkg, os.R_OK):
                self.cached_envs[fn_id] = pkg
                logger.debug("Cached package for %s found at %s", fn_name, pkg)
                return pkg
            (fd, tarball) = tempfile.mkstemp(dir=pkg_dir, prefix='.tmp', suffix='.tar.gz')
            os.close(fd)
            logger.info("Creating dependency package for %s", fn_name)
            logger.debug("Writing deps for %s to %s", fn_name, tarball)
            subprocess.run([self._package_create_script, spec.name, tarball], stdout=subprocess.DEVNULL, check=True)
            logger.debug("Done with conda-pack; moving %s to %s", tarball, pkg)
            os.rename(tarball, pkg)
            self.cached_envs[fn_id] = pkg
            return pkg

    def initialize_scaling(self):
        """ Compose the launch command and call scale out

        Scales the workers to the appropriate nodes with provider
        """
        # Start scaling in/out
        logger.debug("Starting TaskVineExecutor with provider: %s", self.provider)
        self._worker_command = self._construct_worker_command()
        self._patch_providers()

    @property
    def outstanding(self) -> int:
        """Count the number of outstanding tasks."""
        logger.debug(f"Counted {self._outstanding_tasks} outstanding tasks")
        return self._outstanding_tasks

    @property
    def workers_per_node(self) -> Union[int, float]:
        return 1

    def shutdown(self, *args, **kwargs):
        """Shutdown the executor. Sets flag to cancel the submit process and
        collector thread, which shuts down the TaskVine system submission.
        """
        logger.debug("TaskVine shutdown started")
        self._should_stop.set()

        # Join all processes before exiting
        logger.debug("Joining on submit process")
        self._submit_process.join()
        self._submit_process.close()
        logger.debug("Joining on collector thread")
        self._collector_thread.join()
        if self.worker_launch_method == 'factory':
            logger.debug("Joining on factory process")
            self._factory_process.join()
            self._factory_process.close()

        # Shutdown multiprocessing queues
        self._ready_task_queue.close()
        self._ready_task_queue.join_thread()
        self._finished_task_queue.close()
        self._finished_task_queue.join_thread()

        super().shutdown()

        logger.debug("TaskVine shutdown completed")

    @wrap_with_logs
    def _collect_taskvine_results(self):
        """Sets the values of tasks' futures completed by taskvine.
        """
        logger.debug("Starting Collector Thread")
        try:
            while not self._should_stop.is_set():
                if not self._submit_process.is_alive():
                    raise ExecutorError(self, "taskvine Submit Process is not alive")

                # Get the result message from the _finished_task_queue
                try:
                    task_report = self._finished_task_queue.get(timeout=1)
                except queue.Empty:
                    continue

                # Obtain the future from the tasks dictionary
                with self._tasks_lock:
                    future = self.tasks.pop(task_report.executor_id)

                logger.debug(f'Updating Future for Parsl Task: {task_report.executor_id}. '
                             f'Task {task_report.executor_id} has result_received set to {task_report.result_received}')
                if task_report.result_received:
                    try:
                        with open(task_report.result_file, 'rb') as f_in:
                            result = deserialize(f_in.read())
                    except Exception as e:
                        logger.error(f'Cannot load result from result file {task_report.result_file}. Exception: {e}')
                        ex = TaskVineTaskFailure('Cannot load result from result file', None)
                        ex.__cause__ = e
                        future.set_exception(ex)
                    else:
                        if isinstance(result, Exception):
                            ex = TaskVineTaskFailure('Task execution raises an exception', result)
                            ex.__cause__ = result
                            future.set_exception(ex)
                        else:
                            future.set_result(result)
                else:
                    # If there are no results, then the task failed according to one of
                    # taskvine modes, such as resource exhaustion.
                    ex = TaskVineTaskFailure(task_report.reason, None)
                    future.set_exception(ex)

                # decrement outstanding task counter
                with self._outstanding_tasks_lock:
                    self._outstanding_tasks -= 1
        finally:
            logger.debug(f"Marking all {self.outstanding} outstanding tasks as failed")
            logger.debug("Acquiring tasks_lock")
            with self._tasks_lock:
                logger.debug("Acquired tasks_lock")
                # set exception for tasks waiting for results that taskvine did not execute
                for fu in self.tasks.values():
                    if not fu.done():
                        fu.set_exception(TaskVineManagerFailure("taskvine executor failed to execute the task."))
        logger.debug("Exiting Collector Thread")



import os

from parsl.config import Config


def fresh_config():
    from parsl.executors.radical import RadicalPilotExecutor, ResourceConfig

    rpex_cfg = ResourceConfig()
    rpex_cfg.worker_type = "MPI"
    rpex_cfg.worker_cores_per_node = 7

    return Config(
            executors=[
                RadicalPilotExecutor(
                    label='RPEXMPI',
                    rpex_cfg=rpex_cfg,
                    bulk_mode=True,
                    resource='local.localhost',
                    runtime=30, cores=8)])



import logging
import time

import pytest

import parsl
from parsl import File, python_app
from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.jobs.states import TERMINAL_STATES, JobState
from parsl.launchers import SingleNodeLauncher
from parsl.providers import LocalProvider

logger = logging.getLogger(__name__)

_max_blocks = 1
_min_blocks = 0


def local_config():
    return Config(
        executors=[
            HighThroughputExecutor(
                heartbeat_period=1,
                heartbeat_threshold=2,
                poll_period=100,
                label="htex_local",
                address="127.0.0.1",
                max_workers_per_node=1,
                encrypted=True,
                launch_cmd="sleep inf",
                provider=LocalProvider(
                    init_blocks=1,
                    max_blocks=_max_blocks,
                    min_blocks=_min_blocks,
                    launcher=SingleNodeLauncher(),
                ),
            )
        ],
        max_idletime=0.5,
        strategy='htex_auto_scale',
        strategy_period=0.1
    )


# see issue #1885 for details of failures of this test.
# at the time of issue #1885 this test was failing frequently
# in CI.
@pytest.mark.local
def test_scaledown_with_register(try_assert):
    dfk = parsl.dfk()
    htex = dfk.executors['htex_local']

    num_managers = len(htex.connected_managers())
    assert num_managers == 0, "Expected 0 managers at start"

    try_assert(lambda: len(htex.status()),
               fail_msg="Expected 1 block at start")

    s = htex.status()
    assert s['0'].state == JobState.RUNNING, "Expected block to be in RUNNING"

    def check_zero_blocks():
        s = htex.status()
        return len(s) == 1 and s['0'].state in TERMINAL_STATES

    try_assert(
        check_zero_blocks,
        fail_msg="Expected 0 blocks after idle scaledown",
        timeout_ms=15000,
    )



from parsl import python_app
from parsl.dataflow.errors import DependencyError


@python_app
def fails():
    raise ValueError("Deliberate failure")


@python_app
def depends(parent):
    return 1


def test_depfail_once():
    """Test the simplest dependency failure case"""
    f1 = fails()
    f2 = depends(f1)

    assert isinstance(f1.exception(), Exception)
    assert not isinstance(f1.exception(), DependencyError)
    assert isinstance(f2.exception(), DependencyError)

    # check that the task ID of the failing task is mentioned
    # in the DependencyError message
    assert ("task " + str(f1.task_record['id'])) in str(f2.exception())


def test_depfail_chain():
    """Test that dependency failures chain"""
    f1 = fails()
    f2 = depends(f1)
    f3 = depends(f2)
    f4 = depends(f3)

    assert isinstance(f1.exception(), Exception)
    assert not isinstance(f1.exception(), DependencyError)
    assert isinstance(f2.exception(), DependencyError)
    assert isinstance(f3.exception(), DependencyError)
    assert isinstance(f4.exception(), DependencyError)


def test_depfail_branches():
    """Test that dependency failures propagate in the
    presence of multiple downstream tasks."""

    f1 = fails()
    f2 = depends(f1)
    f3 = depends(f1)

    assert isinstance(f1.exception(), Exception)
    assert not isinstance(f1.exception(), DependencyError)
    assert isinstance(f2.exception(), DependencyError)
    assert isinstance(f3.exception(), DependencyError)



from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.launchers import MpiRunLauncher
from parsl.providers import PBSProProvider

config = Config(
        executors=[
            HighThroughputExecutor(
                label="Improv_multinode",
                max_workers_per_node=32,
                provider=PBSProProvider(
                    account="YOUR_ALLOCATION_ON_IMPROV",
                    # PBS directives (header lines), for example:
                    # scheduler_options='#PBS -l mem=4gb',
                    scheduler_options='',

                    queue="compute",

                    # Command to be run before starting a worker:
                    # **WARNING** Improv requires an openmpi module to be
                    # loaded for the MpiRunLauncher. Add additional env
                    # load commands to this multiline string.
                    worker_init='''
module load gcc/13.2.0;
module load openmpi/5.0.3-gcc-13.2.0; ''',
                    launcher=MpiRunLauncher(),

                    # number of compute nodes allocated for each block
                    nodes_per_block=2,
                    walltime='00:10:00'
                ),
            ),
        ],
)



import logging
import os
import re
import time

import typeguard

from parsl.jobs.states import JobState, JobStatus
from parsl.launchers import SingleNodeLauncher
from parsl.launchers.base import Launcher
from parsl.providers.cluster_provider import ClusterProvider
from parsl.providers.condor.template import template_string
from parsl.providers.errors import ScaleOutFailed
from parsl.utils import RepresentationMixin

logger = logging.getLogger(__name__)

from typing import Dict, List, Optional

# See http://pages.cs.wisc.edu/~adesmet/status.html
translate_table = {
    '1': JobState.PENDING,
    '2': JobState.RUNNING,
    '3': JobState.CANCELLED,
    '4': JobState.COMPLETED,
    '5': JobState.FAILED,
    '6': JobState.FAILED,
}


class CondorProvider(RepresentationMixin, ClusterProvider):
    """HTCondor Execution Provider.

    Parameters
    ----------
    nodes_per_block : int
        Nodes to provision per block.
    cores_per_slot : int
        Specify the number of cores to provision per slot. If set to None, executors
        will assume all cores on the node are available for computation. Default is None.
    mem_per_slot : float
        Specify the real memory to provision per slot in GB. If set to None, no
        explicit request to the scheduler will be made. Default is None.
    init_blocks : int
        Number of blocks to provision at time of initialization
    min_blocks : int
        Minimum number of blocks to maintain
    max_blocks : int
        Maximum number of blocks to maintain.
    parallelism : float
        Ratio of provisioned task slots to active tasks. A parallelism value of 1 represents aggressive
        scaling where as many resources as possible are used; parallelism close to 0 represents
        the opposite situation in which as few resources as possible (i.e., min_blocks) are used.
    environment : dict of str
        A dictionary of environmant variable name and value pairs which will be set before
        running a task.
    project : str
        Project which the job will be charged against
    scheduler_options : str
        String to add specific condor attributes to the HTCondor submit script.
    transfer_input_files : list(str)
        List of strings of paths to additional files or directories to transfer to the job
    worker_init : str
        Command to be run before starting a worker.
    requirements : str
        Condor requirements.
    launcher : Launcher
        Launcher for this provider. Possible launchers include
        :class:`~parsl.launchers.SingleNodeLauncher` (the default),
    cmd_timeout : int
        Timeout for commands made to the scheduler in seconds.
    cmd_chunk_size : int
        Calls to the scheduler will be made for chunks of blocks with this size.
    """
    @typeguard.typechecked
    def __init__(self,
                 nodes_per_block: int = 1,
                 cores_per_slot: Optional[int] = None,
                 mem_per_slot: Optional[float] = None,
                 init_blocks: int = 1,
                 min_blocks: int = 0,
                 max_blocks: int = 1,
                 parallelism: float = 1,
                 environment: Optional[Dict[str, str]] = None,
                 project: str = '',
                 scheduler_options: str = '',
                 transfer_input_files: List[str] = [],
                 walltime: str = "00:10:00",
                 worker_init: str = '',
                 launcher: Launcher = SingleNodeLauncher(),
                 requirements: str = '',
                 cmd_timeout: int = 60,
                 cmd_chunk_size: int = 100) -> None:

        label = 'condor'
        super().__init__(label,
                         nodes_per_block,
                         init_blocks,
                         min_blocks,
                         max_blocks,
                         parallelism,
                         walltime,
                         launcher,
                         cmd_timeout=cmd_timeout)
        self.cores_per_slot = cores_per_slot
        self.mem_per_slot = mem_per_slot
        self.cmd_chunk_size = cmd_chunk_size

        # To Parsl, Condor slots should be treated equivalently to nodes
        self.cores_per_node = cores_per_slot
        self.mem_per_node = mem_per_slot

        self.environment = environment if environment is not None else {}
        for key, value in self.environment.items():
            # To escape literal quote marks, double them
            # See: http://research.cs.wisc.edu/htcondor/manual/v8.6/condor_submit.html
            try:
                self.environment[key] = "'{}'".format(value.replace("'", '"').replace('"', '""'))
            except AttributeError:
                pass

        self.project = project
        self.scheduler_options = scheduler_options + '\n'
        self.worker_init = worker_init + '\n'
        self.requirements = requirements
        self.transfer_input_files = transfer_input_files

    def _status(self):
        """Update the resource dictionary with job statuses."""

        for job_id_chunk in _chunker(self.resources.keys(), self.cmd_chunk_size):
            job_id_list = ' '.join(job_id_chunk)
            cmd = "condor_q {0} -af:jr JobStatus".format(job_id_list)
            retcode, stdout, stderr = self.execute_wait(cmd)
            """
            Example output:

            $ condor_q 34524642.0 34524643.0 -af:jr JobStatus
            34524642.0 2
            34524643.0 1
            """

            for line in stdout.splitlines():
                parts = line.strip().split()
                job_id = parts[0]
                state = translate_table.get(parts[1], JobState.UNKNOWN)
                if job_id in self.resources:
                    self.resources[job_id]['status'] = JobStatus(state)

    def status(self, job_ids):
        """Get the status of a list of jobs identified by their ids.

        Parameters
        ----------
        job_ids : list of int
            Identifiers of jobs for which the status will be returned.

        Returns
        -------
        List of int
            Status codes for the requested jobs.

        """
        self._status()
        return [self.resources[jid]['status'] for jid in job_ids]

    def submit(self, command, tasks_per_node, job_name="parsl.condor"):
        """Submits the command onto an Local Resource Manager job.

        example file with the complex case of multiple submits per job:
            Universe =vanilla
            output = out.$(Cluster).$(Process)
            error = err.$(Cluster).$(Process)
            log = log.$(Cluster)
            leave_in_queue = true
            executable = test.sh
            queue 5
            executable = foo
            queue 1

        $ condor_submit test.sub
        Submitting job(s)......
        5 job(s) submitted to cluster 118907.
        1 job(s) submitted to cluster 118908.

        Parameters
        ----------
        command : str
            Command to execute
        job_name : str
            Job name prefix.
        tasks_per_node : int
            command invocations to be launched per node
        Returns
        -------
        None or str
            None if at capacity and cannot provision more; otherwise the identifier for the job.
        """

        logger.debug("Attempting to launch")

        job_name = "parsl.{0}.{1}".format(job_name, time.time())

        scheduler_options = self.scheduler_options
        worker_init = self.worker_init
        if self.mem_per_slot is not None:
            scheduler_options += 'RequestMemory = {}\n'.format(self.mem_per_slot * 1024)
            worker_init += 'export PARSL_MEMORY_GB={}\n'.format(self.mem_per_slot)
        if self.cores_per_slot is not None:
            scheduler_options += 'RequestCpus = {}\n'.format(self.cores_per_slot)
            worker_init += 'export PARSL_CORES={}\n'.format(self.cores_per_slot)

        script_path = "{0}/{1}.submit".format(self.script_dir, job_name)
        script_path = os.path.abspath(script_path)
        userscript_path = "{0}/{1}.script".format(self.script_dir, job_name)
        userscript_path = os.path.abspath(userscript_path)

        self.environment["JOBNAME"] = "'{}'".format(job_name)

        job_config = {}
        job_config["job_name"] = job_name
        job_config["submit_script_dir"] = self.script_dir
        job_config["project"] = self.project
        job_config["nodes"] = self.nodes_per_block
        job_config["scheduler_options"] = scheduler_options
        job_config["worker_init"] = worker_init
        job_config["user_script"] = command
        job_config["tasks_per_node"] = tasks_per_node
        job_config["requirements"] = self.requirements
        job_config["environment"] = ' '.join(['{}={}'.format(key, value) for key, value in self.environment.items()])

        # Move the user script
        # This is where the command should be wrapped by the launchers.
        wrapped_command = self.launcher(command,
                                        tasks_per_node,
                                        self.nodes_per_block)

        with open(userscript_path, 'w') as f:
            f.write(job_config["worker_init"] + '\n' + wrapped_command)

        the_input_files = [userscript_path] + self.transfer_input_files
        job_config["input_files"] = ','.join(the_input_files)
        job_config["job_script"] = os.path.basename(userscript_path)

        # Construct and move the submit script
        self._write_submit_script(template_string, script_path, job_name, job_config)

        cmd = "condor_submit {0}".format(script_path)
        try:
            retcode, stdout, stderr = self.execute_wait(cmd)
        except Exception as e:
            raise ScaleOutFailed(self.label, str(e))

        job_id = []

        if retcode == 0:
            for line in stdout.split('\n'):
                if re.match('^[0-9]', line) is not None:
                    cluster = line.split(" ")[5]
                    # We know the first job id ("process" in condor terms) within a
                    # cluster is 0 and we know the total number of jobs from
                    # condor_submit, so we use some list comprehensions to expand
                    # the condor_submit output into job IDs
                    # e.g., ['118907.0', '118907.1', '118907.2', '118907.3', '118907.4', '118908.0']
                    processes = [str(x) for x in range(0, int(line[0]))]
                    job_id += [cluster + process for process in processes]

            self._add_resource(job_id)
            return job_id[0]
        else:
            message = "Command '{}' failed with return code {}".format(cmd, retcode)
            message += " and standard output '{}'".format(stdout.strip()) if stdout is not None else ''
            message += " and standard error '{}'".format(stderr.strip()) if stderr is not None else ''
            raise ScaleOutFailed(self.label, message)

    def cancel(self, job_ids):
        """Cancels the jobs specified by a list of job IDs.

        Parameters
        ----------
        job_ids : list of str
            The job IDs to cancel.

        Returns
        -------
        list of bool
            Each entry in the list will be True if the job is cancelled succesfully, otherwise False.
        """
        rets = []
        for job_id_chunk in _chunker(job_ids, self.cmd_chunk_size):
            job_id_list = ' '.join(job_id_chunk)
            cmd = "condor_rm {0}; condor_rm -forcex {0}".format(job_id_list)
            logger.debug("Attempting removal of jobs : {0}".format(cmd))
            retcode, stdout, stderr = self.execute_wait(cmd)
            if retcode == 0:
                for jid in job_id_chunk:
                    if jid in self.resources:
                        self.resources[jid]['status'] = JobStatus(JobState.CANCELLED)
                rets.extend([True for i in job_id_chunk])
            else:
                rets.extend([False for i in job_id_chunk])

        return rets

    def _add_resource(self, job_id):
        for jid in job_id:
            self.resources[jid] = {'status': JobStatus(JobState.PENDING), 'size': 1}
        return True

    @property
    def status_polling_interval(self):
        return 60


# see https://stackoverflow.com/a/18243990
def _chunker(seq, size):
    res = []
    for el in seq:
        res.append(el)
        if len(res) == size:
            yield res
            res = []
    if res:
        yield res


if __name__ == "__main__":

    print("None")



import logging
import os
import pathlib
import random
import shutil
import socket
import subprocess
import tempfile
import threading
import time

import pytest

from parsl.jobs.states import JobState
from parsl.launchers import SingleNodeLauncher
from parsl.providers import LocalProvider

logger = logging.getLogger(__name__)


def _run_tests(p: LocalProvider):
    status = _run(p, '/bin/true')
    assert status.message is None
    assert status.state == JobState.COMPLETED
    assert status.exit_code == 0
    assert status.stdout == ''
    assert status.stderr == ''

    status = _run(p, '/bin/true', np=2)
    assert status.message is None
    assert status.state == JobState.COMPLETED
    assert status.exit_code == 0
    assert status.stdout == ''
    assert status.stderr == ''

    status = _run(p, '/bin/false')
    assert status.state == JobState.FAILED
    assert status.exit_code != 0
    assert status.stdout == ''
    assert status.stderr == ''

    status = _run(p, '/bin/false', np=2)
    assert status.state == JobState.FAILED
    assert status.exit_code != 0
    assert status.stdout == ''
    assert status.stderr == ''

    status = _run(p, '/bin/echo -n magic')
    assert status.state == JobState.COMPLETED
    assert status.exit_code == 0
    assert status.stdout == 'magic'
    assert status.stderr == ''

    status = _run(p, '/bin/echo -n magic 1>&2')
    assert status.state == JobState.COMPLETED
    assert status.exit_code == 0
    assert status.stdout == ''
    assert status.stderr == 'magic'


@pytest.mark.local
def test_local_channel():
    with tempfile.TemporaryDirectory() as script_dir:
        script_dir = tempfile.mkdtemp()
        p = LocalProvider(launcher=SingleNodeLauncher(debug=False))
        p.script_dir = script_dir
        _run_tests(p)


def _run(p: LocalProvider, command: str, np: int = 1):
    id = p.submit(command, np)
    return _wait(p, id)


def _wait(p: LocalProvider, id: object):
    status = p.status([id])[0]
    while not status.terminal:
        time.sleep(0.1)
        status = p.status([id])[0]
    return status



import sys
import traceback

from parsl.app.errors import RemoteExceptionWrapper
from parsl.data_provider.files import File
from parsl.serialize import deserialize, serialize
from parsl.utils import get_std_fname_mode

# This scripts executes a parsl function which is pickled in 4 files:
#
# exec_parsl_function.py map_file function_file argument_file result_file
#
# map_file: Contains a pickled dictionary that indicates which local_paths the
#           parsl Files should take.
#
# function_file: Contains a pickle parsl function. Function might be serialized in advance.
# See @parsl.serialize.concretes.py
#
# argument_file: Contains the serialized arguments to the function call.
#
# result_file: A file path, whose content will contain the result of the function, including any
#              exception generated. Exceptions will be wrapped with RemoteExceptionWrapper.
#
# Exit codes:
# 0: The function was evaluated to completion. The result or any exception
#    wrapped with RemoteExceptionWrapper were written to result_file.
# anything else: There was an error that prevented writing to the result file altogether.
#                The exit code corresponds to whatever the python interpreter gives.
#


def dump_result_to_file(result_file: str, result):
    """ Dump a result to the given result file."""
    with open(result_file, "wb") as f_out:
        f_out.write(serialize(result))


def remap_location(mapping, parsl_file):
    """ Remap files from local name (on manager) to remote name (on worker)."""
    if not isinstance(parsl_file, File):
        return
    # Below we rewrite .local_path when scheme != file only when the local_name
    # was given by the main parsl process.  This is the case when scheme !=
    # 'file' but .local_path (via filepath) is in mapping.
    if parsl_file.scheme == 'file' or parsl_file.local_path:
        master_location = parsl_file.filepath
        if master_location in mapping:
            parsl_file.local_path = mapping[master_location]


def remap_list_of_files(mapping, maybe_files):
    """ Remap a list of potential files."""
    for maybe_file in maybe_files:
        remap_location(mapping, maybe_file)


def remap_all_files(mapping, fn_args, fn_kwargs):
    # remap any positional argument given to the function that looks like a
    # File
    remap_list_of_files(mapping, fn_args)

    # remap any keyword argument in the same way, but we need to treat
    # "inputs" and "outputs" specially because they are lists, and
    # "stdout" and "stderr", because they are not File's.
    for kwarg, maybe_file in fn_kwargs.items():
        if kwarg in ["inputs", "outputs"]:
            remap_list_of_files(mapping, maybe_file)
        if kwarg in ["stdout", "stderr"]:
            if maybe_file:
                (fname, mode) = get_std_fname_mode(kwarg, maybe_file)
                if fname in mapping:
                    fn_kwargs[kwarg] = (mapping[fname], mode)
        else:
            # Treat anything else as a possible File to be remapped.
            remap_location(mapping, maybe_file)


def unpack_object_from_file(path):
    with open(path, 'rb') as f:
        obj = deserialize(f.read())
    return obj


def encode_function(user_namespace, fn, fn_name, fn_args, fn_kwargs):
    """ Register the given function to the given namespace."""
    # Returns a tuple (code, result_name)
    # code can be exec in the user_namespace to produce result_name.
    prefix = "parsl_"
    args_name = prefix + "args"
    kwargs_name = prefix + "kwargs"
    result_name = prefix + "result"

    # Add variables to the namespace to make function call
    user_namespace.update({args_name: fn_args,
                           kwargs_name: fn_kwargs,
                           result_name: result_name})

    if callable(fn):
        code = encode_byte_code_function(user_namespace, fn, fn_name, args_name, kwargs_name, result_name)
    else:
        raise ValueError("Function object does not look like a function.")

    return (code, result_name)


def encode_byte_code_function(user_namespace, fn, fn_name, args_name, kwargs_name, result_name):
    user_namespace.update({fn_name: fn})
    code = "{0} = {1}(*{2}, **{3})".format(result_name, fn_name, args_name, kwargs_name)
    return code


def load_function(map_file, function_file, argument_file):
    # Decodes the function and its file arguments to be executed into
    # function_code, and updates a user namespace with the function name and
    # the variable named `result_name`. When the function is executed, its result
    # will be stored in this variable in the user namespace.
    # Returns (namespace, function_code, result_name)

    fn = unpack_object_from_file(function_file)
    args_dict = unpack_object_from_file(argument_file)
    fn_args = args_dict['args']
    fn_kwargs = args_dict['kwargs']
    fn_name = 'parsl_tmp_func_name'

    mapping = unpack_object_from_file(map_file)
    remap_all_files(mapping, fn_args, fn_kwargs)

    # Create the namespace to isolate the function execution.
    user_ns = locals()
    user_ns.update({'__builtins__': __builtins__})

    (code, result_name) = encode_function(user_ns, fn, fn_name, fn_args, fn_kwargs)

    return (user_ns, code, result_name)


def execute_function(namespace, function_code, result_name):
    # On executing the function inside the namespace, its result will be in a
    # variable named result_name.
    exec(function_code, namespace, namespace)
    result = namespace.get(result_name)

    return result


def run(map_file, function_file, argument_file, result_file):
    try:
        try:
            (namespace, function_code, result_name) = load_function(map_file, function_file, argument_file)
        except Exception:
            print("There was an error setting up the function for execution.")
            raise

        try:
            result = execute_function(namespace, function_code, result_name)
        except Exception:
            print("There was an error executing the function.")
            raise
    except Exception:
        traceback.print_exc()
        result = RemoteExceptionWrapper(*sys.exc_info())

    # Write out function result to the result file
    try:
        dump_result_to_file(result_file, result)
    except Exception:
        print("Could not write to result file.")
        traceback.print_exc()
        sys.exit(1)


if __name__ == "__main__":
    # parse the four required command line arguments:
    # map_file: contains a pickled dictionary to map original names to
    #           names at the execution site.
    # function_file: contains the pickled parsl function to execute.
    # argument_file: contains the pickled arguments to the function call.
    # result_file: any output (including exceptions) will be written to
    #              this file.
    try:
        (map_file, function_file, argument_file, result_file) = sys.argv[1:]
    except ValueError:
        print("Usage:\n\t{} function argument result mapping\n".format(sys.argv[0]))
        raise
    run(map_file, function_file, argument_file, result_file)



import hashlib
import hmac
import logging
import pickle
import secrets
import socket
from multiprocessing.queues import Queue
from typing import Optional

from parsl.monitoring.radios.base import (
    MonitoringRadioReceiver,
    MonitoringRadioSender,
    RadioConfig,
)
from parsl.monitoring.radios.udp_router import start_udp_receiver

logger = logging.getLogger(__name__)


class UDPRadio(RadioConfig):
    def __init__(self, *, port: Optional[int] = None, atexit_timeout: int = 3, address: str, debug: bool = False, hmac_digest: str = 'sha512'):
        self.port = port
        self.atexit_timeout = atexit_timeout
        self.address = address
        self.debug = debug
        self.hmac_digest = hmac_digest

    def create_sender(self) -> MonitoringRadioSender:
        assert self.port is not None, "self.port should have been initialized by create_receiver"
        return UDPRadioSender(self.address, self.port, self.hmac_key, self.hmac_digest)

    def create_receiver(self, run_dir: str, resource_msgs: Queue) -> MonitoringRadioReceiver:
        # RFC 2104 section 2 recommends that the key length be at
        # least as long as the hash output (64 bytes in the case of SHA512).
        # RFC 2014 section 3 talks about periodic key refreshing. This key is
        # not refreshed inside a workflow run, but each separate workflow run
        # uses a new key.
        keysize = hashlib.new(self.hmac_digest).digest_size
        self.hmac_key = secrets.token_bytes(keysize)

        udp_receiver = start_udp_receiver(logdir=run_dir,
                                          monitoring_messages=resource_msgs,
                                          port=self.port,
                                          debug=self.debug,
                                          atexit_timeout=self.atexit_timeout,
                                          hmac_key=self.hmac_key,
                                          hmac_digest=self.hmac_digest
                                          )
        self.port = udp_receiver.port
        return udp_receiver


class UDPRadioSender(MonitoringRadioSender):

    def __init__(self, address: str, port: int, hmac_key: bytes, hmac_digest: str, *, timeout: int = 10) -> None:
        self.sock_timeout = timeout
        self.address = address
        self.port = port
        self.hmac_key = hmac_key
        self.hmac_digest = hmac_digest

        self.sock = socket.socket(socket.AF_INET,
                                  socket.SOCK_DGRAM,
                                  socket.IPPROTO_UDP)  # UDP
        self.sock.settimeout(self.sock_timeout)

    def send(self, message: object) -> None:
        """ Sends a message to the UDP receiver

        Parameter
        ---------

        message: object
            Arbitrary pickle-able object that is to be sent

        Returns:
            None
        """
        logger.info("Starting UDP radio message send")
        try:
            data = pickle.dumps(message)
            origin_hmac = hmac.digest(self.hmac_key, data, self.hmac_digest)
            buffer = origin_hmac + data
        except Exception:
            logging.exception("Exception during pickling", exc_info=True)
            return

        try:
            self.sock.sendto(buffer, (self.address, self.port))
        except socket.timeout:
            logging.error("Could not send message within timeout limit")
            return
        logger.info("Normal ending for UDP radio message send")
        return



import pytest

import parsl
from parsl.app.app import python_app
from parsl.tests.configs.htex_local import fresh_config


def local_config():
    config = fresh_config()
    config.executors[0].poll_period = 1
    config.executors[0].max_workers_per_node = 1
    config.executors[0].launch_cmd = "executable_that_hopefully_does_not_exist_1030509.py"
    return config


@python_app
def dummy():
    pass


@pytest.mark.local
def test_that_it_fails():
    x = dummy()
    failed = False
    try:
        x.result()
    except Exception as ex:
        print(ex)
        failed = True
    if not failed:
        raise Exception("The app somehow ran without a valid worker")

    assert parsl.dfk().config.executors[0]._executor_bad_state.is_set()



import os

import pytest

import parsl
from parsl import python_app
from parsl.tests.configs.local_threads import fresh_config


def local_config():
    config = fresh_config()
    config.checkpoint_mode = "manual"
    return config


@python_app(cache=True)
def uuid_app():
    import uuid
    return uuid.uuid4()


@pytest.mark.local
def test_initial_checkpoint_write():
    """1. Launch a few apps and write the checkpoint once a few have completed
    """
    uuid_app().result()

    cpt_dir = parsl.dfk().checkpoint()

    cptpath = cpt_dir + '/tasks.pkl'
    assert os.path.exists(cptpath), f"Tasks checkpoint missing: {cptpath}"



import logging
from abc import abstractmethod
from string import Template

from parsl.launchers.base import Launcher
from parsl.launchers.errors import BadLauncher
from parsl.providers.base import ExecutionProvider
from parsl.providers.errors import SchedulerMissingArgs, ScriptPathError
from parsl.utils import execute_wait

logger = logging.getLogger(__name__)


class ClusterProvider(ExecutionProvider):
    """ This class defines behavior common to all cluster/supercompute-style scheduler systems.

    Parameters
    ----------
      label : str
        Label for this provider.
      walltime : str
        Walltime requested per block in HH:MM:SS.
      launcher : Launcher
        Launcher for this provider.
      cmd_timeout : int
        Timeout for commands made to the scheduler in seconds

    .. code:: python

                                +------------------
                                |
          script_string ------->|  submit
               id      <--------|---+
                                |
          [ ids ]       ------->|  status
          [statuses]   <--------|----+
                                |
          [ ids ]       ------->|  cancel
          [cancel]     <--------|----+
                                |
                                +-------------------
    """

    def __init__(self,
                 label,
                 nodes_per_block,
                 init_blocks,
                 min_blocks,
                 max_blocks,
                 parallelism,
                 walltime,
                 launcher,
                 cmd_timeout=10):

        self._label = label
        self.nodes_per_block = nodes_per_block
        self.init_blocks = init_blocks
        self.min_blocks = min_blocks
        self.max_blocks = max_blocks
        self.parallelism = parallelism
        self.launcher = launcher
        self.walltime = walltime
        self.cmd_timeout = cmd_timeout
        if not isinstance(self.launcher, Launcher):
            raise BadLauncher(self.launcher)

        self.script_dir = None

        # Dictionary that keeps track of jobs, keyed on job_id
        self.resources = {}

    def execute_wait(self, cmd, timeout=None):
        t = self.cmd_timeout
        if timeout is not None:
            t = timeout
        return execute_wait(cmd, t)

    def _write_submit_script(self, template, script_filename, job_name, configs):
        """Generate submit script and write it to a file.

        Args:
              - template (string) : The template string to be used for the writing submit script
              - script_filename (string) : Name of the submit script
              - job_name (string) : job name
              - configs (dict) : configs that get pushed into the template

        Returns:
              - None

        Raises:
              SchedulerMissingArgs : If template is missing args
              ScriptPathError : Unable to write submit script out
        """

        try:
            submit_script = Template(template).substitute(jobname=job_name, **configs)
            with open(script_filename, 'w') as f:
                f.write(submit_script)

        except KeyError as e:
            logger.error("Missing keys for submit script : %s", e)
            raise SchedulerMissingArgs(e.args, self.label)

        except IOError as e:
            logger.error("Failed writing to submit script: %s", script_filename)
            raise ScriptPathError(script_filename, e)
        except Exception as e:
            print("Template : ", template)
            print("Args : ", job_name)
            print("Kwargs : ", configs)
            logger.error("Uncategorized error: %s", e)
            raise e

    @abstractmethod
    def _status(self):
        pass

    def status(self, job_ids):
        """ Get the status of a list of jobs identified by the job identifiers
        returned from the submit request.

        Args:
             - job_ids (list) : A list of job identifiers

        Returns:
             - A list of JobStatus objects corresponding to each job_id in the job_ids list.

        Raises:
             - ExecutionProviderException or its subclasses

        """
        if job_ids:
            self._status()
        return [self.resources[jid]['status'] for jid in job_ids]

    @property
    def label(self):
        return self._label



import parsl


@parsl.python_app
def k(x):
    return x


@parsl.python_app
def mul(left, right):
    return left * right


def test_diamond_dag():
    a = k(2)
    b = mul(a, 3)
    c = mul(a, 5)
    d = mul(b, c)
    e = k(d)

    assert e.result() == (2 * 3) * (2 * 5)



import hashlib
import logging
import os
import queue
import shutil
import subprocess
import uuid

import parsl
from parsl.executors.taskvine import exec_parsl_function
from parsl.executors.taskvine.utils import VineTaskToParsl, run_parsl_function
from parsl.process_loggers import wrap_with_logs
from parsl.utils import setproctitle

try:
    from ndcctools.taskvine import FunctionCall, Manager, Task, cvine
    from ndcctools.taskvine.cvine import (
        VINE_ALLOCATION_MODE_EXHAUSTIVE_BUCKETING,
        VINE_ALLOCATION_MODE_MAX,
        VINE_ALLOCATION_MODE_MAX_THROUGHPUT,
    )
except ImportError:
    _taskvine_enabled = False
else:
    _taskvine_enabled = True

logger = logging.getLogger(__name__)


def _set_manager_attributes(m, config):
    """ Set various manager global attributes."""
    if config.project_password_file:
        m.set_password_file(config.project_password_file)

    # Autolabeling resources require monitoring to be enabled
    if config.autolabel:
        m.enable_monitoring()
        if config.autolabel_window is not None:
            m.tune('category-steady-n-tasks', config.autolabel_window)

    # Specify number of workers to wait for before sending the first task
    if config.wait_for_workers:
        m.tune("wait-for-workers", config.wait_for_workers)

    # Enable peer transfer feature between workers if specified
    if config.enable_peer_transfers:
        m.enable_peer_transfers()
    else:
        m.disable_peer_transfers()

    # Set catalog report to parsl if project name exists
    if m.name:
        m.set_property("framework", "parsl")

    if config.tune_parameters is not None:
        for k, v in config.tune_parameters.items():
            m.tune(k, v)


def _prepare_environment_serverless(manager_config, env_cache_dir, poncho_create_script):
    # Return path to a packaged poncho environment
    poncho_env_path = ''
    if not manager_config.shared_fs:
        if manager_config.env_pack is None:
            raise Exception('TaskVine manager needs env_pack to be specified when running tasks in serverless mode and with no shared_fs')

        poncho_env_path = manager_config.env_pack

        # If a conda environment name or path is given, then use the hash of the headers of
        # all contained packages as the name of the to-be-packaged poncho tarball,
        # and package it if it's not cached.
        if not poncho_env_path.endswith('tar.gz'):
            if os.path.isabs(poncho_env_path):
                conda_env_signature = hashlib.md5(subprocess.check_output(['conda', 'list', '-p', poncho_env_path, '--json'])).hexdigest()
                logger.debug(f'Signature of conda environment at {poncho_env_path}: {conda_env_signature}')
            else:
                conda_env_signature = hashlib.md5(subprocess.check_output(['conda', 'list', '-n', poncho_env_path, '--json'])).hexdigest()
                logger.debug(f'Signature of conda environment named {poncho_env_path}: {conda_env_signature}')

            # If env is cached then use it,
            # else create a new env tarball
            poncho_env_path = os.path.join(env_cache_dir, '.'.join([conda_env_signature, 'tar.gz']))
            if not os.path.isfile(poncho_env_path):
                logger.debug(f'No cached poncho environment. Creating poncho environment for library task at {poncho_env_path}')
                try:
                    subprocess.run([poncho_create_script, manager_config.env_pack, poncho_env_path], stdout=subprocess.DEVNULL, check=True)
                except Exception:
                    logger.error('Cannot create a poncho environment. Removing it.')
                    if os.path.isfile(poncho_env_path):
                        os.remove(poncho_env_path)
                    raise
            else:
                logger.debug(f'Found cached poncho environment at {poncho_env_path}. Reusing it.')
        else:
            logger.debug(f'Use the given poncho environment at {manager_config.env_pack} to setup library task.')
    return poncho_env_path


def _prepare_environment_regular(m, manager_config, t, task, poncho_env_to_file, poncho_create_script):
    # Add environment file to the task if possible
    # Prioritize local poncho environment over global poncho environment
    # (local: use app_pack, global: use env_pack)
    poncho_env_file = None

    # check if env_pack is specified
    if manager_config.env_pack is not None:

        # check if the environment file is not already created
        if manager_config.env_pack not in poncho_env_to_file:

            # if the environment is already packaged as a tarball, then add the file
            # otherwise it is an environment name or path, so create a poncho tarball then add it
            if not manager_config.env_pack.endswith('.tar.gz'):
                env_tarball = str(uuid.uuid4()) + '.tar.gz'
                logger.debug(f'Creating a poncho environment at {env_tarball} from conda environment {manager_config.env_pack}')
                subprocess.run([poncho_create_script, manager_config.env_pack, env_tarball], stdout=subprocess.DEVNULL, check=True)
            else:
                env_tarball = manager_config.env_pack
            poncho_env_file = m.declare_poncho(env_tarball, cache=True, peer_transfer=True)
            poncho_env_to_file[manager_config.env_pack] = poncho_env_file
        else:
            poncho_env_file = poncho_env_to_file[manager_config.env_pack]
            logger.debug(f'Found cached poncho environment for {manager_config.env_pack}. Reusing it.')

    # check if app_pack is used, override if possible
    if task.env_pkg is not None:
        if task.env_pkg not in poncho_env_to_file:
            poncho_env_file = m.declare_poncho(task.env_pkg, cache=True, peer_transfer=True)
            poncho_env_to_file[task.env_pkg] = poncho_env_file
        else:
            poncho_env_file = poncho_env_to_file[task.env_pkg]

    # Add environment to the task
    if poncho_env_file is not None:
        t.add_environment(poncho_env_file)


@wrap_with_logs
def _taskvine_submit_wait(ready_task_queue=None,
                          finished_task_queue=None,
                          should_stop=None,
                          manager_config=None
                          ):
    """Process to handle Parsl app submissions to the TaskVine objects.
    Takes in Parsl functions submitted using submit(), and creates a
    TaskVine task with the appropriate specifications, which is then
    submitted to TaskVine. After tasks are completed, processes the
    exit status and exit code of the task, and sends results to the
    TaskVine collector thread.
    To avoid python's global interpreter lock with taskvine's wait, this
    function should be launched as a process, not as a lightweight thread. This
    means that any communication should be done using the multiprocessing
    module capabilities, rather than shared memory.
    """
    logger.debug("Starting TaskVine Submit/Wait Process")
    setproctitle("parsl: TaskVine submit/wait")

    # Enable debugging flags and create logging file
    if manager_config.vine_log_dir is not None:
        logger.debug("Setting debugging flags and creating logging file at {}".format(manager_config.vine_log_dir))

    # Create TaskVine queue object
    logger.debug("Creating TaskVine Object")
    try:
        logger.debug("Listening on port {}".format(manager_config.port))
        m = Manager(port=manager_config.port,
                    name=manager_config.project_name,
                    run_info_path=manager_config.vine_log_dir)
    except Exception as e:
        logger.error("Unable to create TaskVine object: {}".format(e))
        raise e

    # Specify TaskVine manager attributes
    _set_manager_attributes(m, manager_config)

    # Get parent pid, useful to shutdown this process when its parent, the taskvine
    # executor process, exits.
    orig_ppid = os.getppid()

    result_file_of_task_id = {}  # Mapping executor task id -> result file.

    poncho_env_to_file = {}  # Mapping poncho_env id to File object in TaskVine

    # Mapping of parsl local file name to TaskVine File object
    # dict[str] -> vine File object
    parsl_file_name_to_vine_file = {}

    # Mapping of tasks from vine id to parsl id
    # Dict[str] -> str
    vine_id_to_executor_task_id = {}

    # Find poncho scripts to create and activate an environment tarball
    poncho_create_script = shutil.which("poncho_package_create")

    # Declare helper script as cache-able and peer-transferable
    exec_parsl_function_file = m.declare_file(exec_parsl_function.__file__, cache=True, peer_transfer=True)

    # Flag to make sure library for serverless tasks is declared and installed only once.
    lib_installed = False

    # Create cache dir for environment files
    env_cache_dir = os.path.join(manager_config.vine_log_dir, 'vine-cache', 'vine-poncho-env-cache')
    os.makedirs(env_cache_dir, exist_ok=True)

    logger.debug("Entering main loop of TaskVine manager")

    while not should_stop.is_set():
        # Check if executor process is still running
        ppid = os.getppid()
        if ppid != orig_ppid:
            logger.debug("Executor process is detected to have exited. Exiting..")
            break

        # Submit tasks
        while ready_task_queue.qsize() > 0 or m.empty() and not should_stop.is_set():
            # Obtain task from ready_task_queue
            try:
                task = ready_task_queue.get(timeout=1)
                logger.debug("Removing executor task from queue")
            except queue.Empty:
                logger.debug("Queue is empty")
                continue
            if task.exec_mode == 'regular':
                # Create command string
                launch_cmd = "python3 exec_parsl_function.py {mapping} {function} {argument} {result}"
                if manager_config.init_command != '':
                    launch_cmd = "{init_cmd} " + launch_cmd
                command_str = launch_cmd.format(init_cmd=manager_config.init_command,
                                                mapping=os.path.basename(task.map_file),
                                                function=os.path.basename(task.function_file),
                                                argument=os.path.basename(task.argument_file),
                                                result=os.path.basename(task.result_file))
                logger.debug("Sending executor task {} (mode: regular) with command: {}".format(task.executor_id, command_str))
                try:
                    t = Task(command_str)
                except Exception as e:
                    logger.error("Unable to create executor task (mode:regular): {}".format(e))
                    finished_task_queue.put_nowait(VineTaskToParsl(executor_id=task.executor_id,
                                                                   result_received=False,
                                                                   result_file=None,
                                                                   reason="task could not be created by taskvine",
                                                                   status=-1))
                    continue
            elif task.exec_mode == 'serverless':
                if not lib_installed:
                    # Declare and install common library for serverless tasks.
                    # Library requires an environment setup properly, which is
                    # different from setup of regular tasks.
                    # If shared_fs is True, then no environment preparation is done.
                    # Only the core serverless code is created.
                    poncho_env_path = _prepare_environment_serverless(manager_config, env_cache_dir, poncho_create_script)

                    # Don't automatically add environment so manager can declare and cache the vine file associated with the environment file
                    add_env = False
                    serverless_lib = m.create_library_from_functions('common-parsl-taskvine-lib',
                                                                     run_parsl_function,
                                                                     poncho_env=poncho_env_path,
                                                                     init_command=manager_config.init_command,
                                                                     add_env=add_env,
                                                                     hoisting_modules=[parsl.serialize, run_parsl_function])

                    # Configure the library if provided
                    if manager_config.library_config:
                        lib_cores = manager_config.library_config.get('cores', None)
                        lib_memory = manager_config.library_config.get('memory', None)
                        lib_disk = manager_config.library_config.get('disk', None)
                        lib_slots = manager_config.library_config.get('num_slots', None)
                        if lib_cores:
                            serverless_lib.set_cores(lib_cores)
                        if lib_memory:
                            serverless_lib.set_memory(lib_memory)
                        if lib_disk:
                            serverless_lib.set_disk(lib_disk)
                        if lib_slots:
                            serverless_lib.set_function_slots(lib_slots)

                    if poncho_env_path:
                        serverless_lib_env_file = m.declare_poncho(poncho_env_path, cache=True, peer_transfer=True)
                        serverless_lib.add_environment(serverless_lib_env_file)
                        poncho_env_to_file[manager_config.env_pack] = serverless_lib_env_file
                        logger.debug(f'Created library task using poncho environment at {poncho_env_path}.')
                    else:
                        logger.debug('Created minimal library task with no environment.')

                    m.install_library(serverless_lib)
                    lib_installed = True
                try:
                    # run_parsl_function only needs remote names of map_file, function_file, argument_file,
                    # and result_file, which are simply named map, function, argument, result.
                    # These names are given when these files are declared below.
                    t = FunctionCall('common-parsl-taskvine-lib', run_parsl_function.__name__, 'map', 'function', 'argument', 'result')
                except Exception as e:
                    logger.error("Unable to create executor task (mode:serverless): {}".format(e))
                    finished_task_queue.put_nowait(VineTaskToParsl(executor_id=task.executor_id,
                                                                   result_received=False,
                                                                   result_file=None,
                                                                   reason="task could not be created by taskvine",
                                                                   status=-1))
            else:
                raise Exception(f'Unrecognized task mode {task.exec_mode}. Exiting...')

            # prepare environment for regular tasks if not using shared_fs
            if task.exec_mode == 'regular' and not manager_config.shared_fs:
                _prepare_environment_regular(m, manager_config, t, task, poncho_env_to_file, poncho_create_script)

            t.set_category(task.category)

            # Set autolabel mode
            if manager_config.autolabel:
                if manager_config.autolabel_algorithm == 'max-xput':
                    m.set_category_mode(task.category, VINE_ALLOCATION_MODE_MAX_THROUGHPUT)
                elif manager_config.autolabel_algorithm == 'bucketing':
                    m.set_category_mode(task.category, VINE_ALLOCATION_MODE_EXHAUSTIVE_BUCKETING)
                elif manager_config.autolabel_algorithm == 'max':
                    m.set_category_mode(task.category, VINE_ALLOCATION_MODE_MAX)
                else:
                    logger.warning(f'Unrecognized autolabeling algorithm named {manager_config.autolabel_algorithm} for taskvine manager.')
                    raise Exception(f'Unrecognized autolabeling algorithm named {manager_config.autolabel_algorithm} for taskvine manager.')

            if task.cores is not None:
                t.set_cores(task.cores)
            if task.memory is not None:
                t.set_memory(task.memory)
            if task.disk is not None:
                t.set_disk(task.disk)
            if task.gpus is not None:
                t.set_gpus(task.gpus)
            if task.priority is not None:
                t.set_priority(task.priority)
            if task.running_time_min is not None:
                t.set_time_min(task.running_time_min)

            if manager_config.max_retries is not None:
                logger.debug(f"Specifying max_retries {manager_config.max_retries}")
                t.set_retries(manager_config.max_retries)

            # Specify environment variables for the task
            if manager_config.env_vars is not None:
                for var in manager_config.env_vars:
                    t.set_env_var(str(var), str(manager_config.env_vars[var]))

            if task.exec_mode == 'regular':
                # Add helper files that execute parsl functions on remote nodes
                # only needed to add as file for tasks with 'regular' mode
                t.add_input(exec_parsl_function_file, "exec_parsl_function.py")

            # Declare and add task-specific function, data, and result files to task
            task_function_file = m.declare_file(task.function_file, cache=False, peer_transfer=False)
            t.add_input(task_function_file, "function")

            task_argument_file = m.declare_file(task.argument_file, cache=False, peer_transfer=False)
            t.add_input(task_argument_file, "argument")

            task_map_file = m.declare_file(task.map_file, cache=False, peer_transfer=False)
            t.add_input(task_map_file, "map")

            task_result_file = m.declare_file(task.result_file, cache=False, peer_transfer=False)
            t.add_output(task_result_file, "result")

            result_file_of_task_id[str(task.executor_id)] = task.result_file

            logger.debug("Executor task id: {}".format(task.executor_id))

            # Specify input/output files that need to be staged.
            # Absolute paths are assumed to be in shared filesystem, and thus
            # not staged by taskvine.
            # Files that share the same local path are assumed to be the same
            # and thus use the same Vine File object if detected.
            if not manager_config.shared_fs:
                for spec in task.input_files:
                    if spec.stage:
                        if spec.parsl_name in parsl_file_name_to_vine_file:
                            task_in_file = parsl_file_name_to_vine_file[spec.parsl_name]
                        else:
                            task_in_file = m.declare_file(spec.parsl_name, cache=spec.cache, peer_transfer=True)
                            parsl_file_name_to_vine_file[spec.parsl_name] = task_in_file
                        t.add_input(task_in_file, spec.parsl_name)

                for spec in task.output_files:
                    if spec.stage:
                        if spec.parsl_name in parsl_file_name_to_vine_file:
                            task_out_file = parsl_file_name_to_vine_file[spec.parsl_name]
                        else:
                            task_out_file = m.declare_file(spec.parsl_name, cache=spec.cache, peer_transfer=True)
                            parsl_file_name_to_vine_file[spec.parsl_name] = task_out_file
                        t.add_output(task_out_file, spec.parsl_name)

            # Submit the task to the TaskVine object
            logger.debug("Submitting executor task {}, {} to TaskVine".format(task.executor_id, t))
            try:
                vine_id = m.submit(t)
                logger.debug("Submitted executor task {} to TaskVine".format(task.executor_id))
                vine_id_to_executor_task_id[str(vine_id)] = str(task.executor_id), task.exec_mode
            except Exception as e:
                logger.error("Unable to submit task to taskvine: {}".format(e))
                finished_task_queue.put_nowait(VineTaskToParsl(executor_id=task.executor_id,
                                                               result_received=False,
                                                               result_file=None,
                                                               reason="task could not be submited to taskvine",
                                                               status=-1))
                continue

            logger.debug("Executor task {} submitted as TaskVine task with id {}".format(task.executor_id, vine_id))

        # If the queue is not empty wait on the TaskVine queue for a task
        task_found = True
        while not m.empty() and task_found and not should_stop.is_set():
            # Obtain the task from the queue
            t = m.wait(1)
            if t is None:
                task_found = False
                continue
            logger.debug('Found a task')
            executor_task_id = vine_id_to_executor_task_id[str(t.id)][0]
            vine_id_to_executor_task_id.pop(str(t.id))

            # When a task is found
            result_file = result_file_of_task_id.pop(executor_task_id)

            logger.debug(f"completed executor task info: {executor_task_id}, {t.category}, {t.command}, {t.std_output}")

            # A tasks completes 'succesfully' if it has result file.
            # A check whether the Python object represented using this file can be
            # deserialized happens later in the collector thread of the executor
            # process.
            logger.debug("Looking for result in {}".format(result_file))
            if os.path.exists(result_file):
                logger.debug("Found result in {}".format(result_file))
                finished_task_queue.put_nowait(VineTaskToParsl(executor_id=executor_task_id,
                                                               result_received=True,
                                                               result_file=result_file,
                                                               reason=None,
                                                               status=t.exit_code))
            # If a result file could not be generated, explain the
            # failure according to taskvine error codes.
            else:
                reason = _explain_taskvine_result(t)
                logger.debug("Did not find result in {}".format(result_file))
                logger.debug("Wrapper Script status: {}\nTaskVine Status: {}".format(t.exit_code, t.result))
                logger.debug("Task with executor id {} / vine id {} failed because:\n{}".format(executor_task_id, t.id, reason))
                finished_task_queue.put_nowait(VineTaskToParsl(executor_id=executor_task_id,
                                                               result_received=False,
                                                               result_file=None,
                                                               reason=reason,
                                                               status=t.exit_code))

    logger.debug("Exiting TaskVine Monitoring Process")
    return 0


def _explain_taskvine_result(vine_task):
    """Returns a string with the reason why a task failed according to taskvine."""

    vine_result = vine_task.result
    reason = "taskvine result: "
    if vine_result == cvine.VINE_RESULT_SUCCESS:
        reason += "succesful execution with exit code {}".format(vine_task.return_status)
    elif vine_result == cvine.VINE_RESULT_OUTPUT_MISSING:
        reason += "The result file was not transfered from the worker.\n"
        reason += "This usually means that there is a problem with the python setup,\n"
        reason += "or the wrapper that executes the function."
        reason += "\nTrace:\n" + str(vine_task.output)
    elif vine_result == cvine.VINE_RESULT_INPUT_MISSING:
        reason += "missing input file"
    elif vine_result == cvine.VINE_RESULT_STDOUT_MISSING:
        reason += "stdout has been truncated"
    elif vine_result == cvine.VINE_RESULT_SIGNAL:
        reason += "task terminated with a signal"
    elif vine_result == cvine.VINE_RESULT_RESOURCE_EXHAUSTION:
        reason += "task used more resources than requested"
    elif vine_result == cvine.VINE_RESULT_MAX_END_TIME:
        reason += "task ran past the specified end time"
    elif vine_result == cvine.VINE_RESULT_UNKNOWN:
        reason += "result could not be classified"
    elif vine_result == cvine.VINE_RESULT_FORSAKEN:
        reason += "task failed, but not a task error"
    elif vine_result == cvine.VINE_RESULT_MAX_RETRIES:
        reason += "unable to complete after specified number of retries"
    elif vine_result == cvine.VINE_RESULT_MAX_WALL_TIME:
        reason += "task ran for more than the specified time"
    elif vine_result == cvine.VINE_RESULT_RMONITOR_ERROR:
        reason += "task failed because the monitor did not produce an output"
    elif vine_result == cvine.VINE_RESULT_OUTPUT_TRANSFER_ERROR:
        reason += "task failed because output transfer fails"
    elif vine_result == cvine.VINE_RESULT_FIXED_LOCATION_MISSING:
        reason += "task failed because no worker could satisfy the fixed \n"
        reason += "location input file requirements"
    else:
        reason += "unable to process TaskVine system failure"
    return reason



import pytest

from parsl.utils import RepresentationMixin


class GoodRepr(RepresentationMixin):
    def __init__(self, x, y):
        self.x = x
        self.y = y


class BadRepr(RepresentationMixin):
    """This class incorrectly subclasses RepresentationMixin.
    It does not store the parameter x on self.
    """
    def __init__(self, x, y):
        self.y = y


@pytest.mark.local
def test_repr_good():
    p1 = "parameter 1"
    p2 = "the second parameter"

    # repr should not raise an exception
    r = repr(GoodRepr(p1, p2))

    # representation should contain both values supplied
    # at object creation.
    assert p1 in r
    assert p2 in r


@pytest.mark.local
def test_repr_bad():
    p1 = "parameter 1"
    p2 = "the second parameter"

    # repr should raise an exception
    with pytest.raises(AttributeError):
        repr(BadRepr(p1, p2))


class NonValidatingRepresentationMixin(RepresentationMixin):
    """This will override the process level RepresentationMixin which can
    be set to validating mode by pytest fixtures"""
    _validate_repr = False


class BadReprNonValidating(NonValidatingRepresentationMixin):
    """This class incorrectly subclasses RepresentationMixin.
    It does not store the parameter x on self.
    """
    def __init__(self, x, y):
        self.y = y


@pytest.mark.local
def test_repr_bad_unvalidated():
    p1 = "parameter 1"
    p2 = "the second parameter"

    # repr should not raise an exception
    r = repr(BadReprNonValidating(p1, p2))
    # parameter 2 should be found in the representation, but not
    # parameter 1
    assert p1 not in r
    assert p2 in r



"""Tests related to Parsl workers being able to access their worker ID"""

import pytest

from parsl import python_app
from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.providers import LocalProvider


def local_config():
    return Config(
        executors=[
            HighThroughputExecutor(
                label="htex_Local",
                worker_debug=True,
                max_workers_per_node=4,
                encrypted=True,
                provider=LocalProvider(
                    init_blocks=1,
                    max_blocks=1,
                ),
            )
        ],
        strategy='none',
    )


@python_app
def get_worker_info():
    import os
    rank = int(os.environ['PARSL_WORKER_RANK'])
    size = int(os.environ['PARSL_WORKER_COUNT'])
    pool_id = os.environ['PARSL_WORKER_POOL_ID']
    return rank, size, pool_id


@pytest.mark.local
def test_htex():
    worker_info = [get_worker_info() for _ in range(4)]
    worker_ids, worker_size, pool_info = zip(*[r.result() for r in worker_info])
    assert len(set(worker_info)) > 1  # Tasks should run on >1 worker
    assert len(set(worker_size)) == 1  # All workers have same pool size
    assert len(set(pool_info)) == 1  # All from the same pool



# this file contains some tests of passing complicated
# callables as parameters to apps, mainly to check that
# they are serialised correctly with serialization based
# executors.

import importlib
import pathlib
from functools import partial

import parsl


@parsl.python_app
def app(x):
    return True


def test_check_base_app():
    app(0).result()


def somefunc(*args):
    pass


def test_check_this_module_function():
    app(somefunc).result()


def test_check_this_module_function_partial():
    app(partial(somefunc, 1)).result()


def test_check_import_module_function():
    from parsl.tests.callables_helper import some_aux_func
    app(some_aux_func).result()


def test_check_import_module_function_partial():
    from parsl.tests.callables_helper import some_aux_func
    app(partial(some_aux_func, 1)).result()


def test_check_importlib_file_function():
    helper_path = pathlib.Path(__file__).parent / "callables_helper.py"
    spec = importlib.util.spec_from_file_location("dynamically_loaded_module", helper_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    some_aux_func = module.some_aux_func
    app(some_aux_func).result()


def test_check_importlib_file_function_partial():
    helper_path = pathlib.Path(__file__).parent / "callables_helper.py"
    spec = importlib.util.spec_from_file_location("dynamically_loaded_module", helper_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    some_aux_func = module.some_aux_func
    app(partial(some_aux_func, 1)).result()


def test_check_importlib_module_function():
    module = importlib.import_module("parsl.tests.callables_helper")
    some_aux_func = module.some_aux_func
    app(some_aux_func).result()


def test_check_importlib_module_partial():
    module = importlib.import_module("parsl.tests.callables_helper")
    some_aux_func = module.some_aux_func
    app(partial(some_aux_func, 1)).result()



import argparse

import parsl
from parsl.app.app import python_app


@python_app(cache=True)
def random_uuid(x, cache=True):
    import uuid
    return str(uuid.uuid4())


def test_python_memoization(n=2):
    """Testing python memoization."""
    x = random_uuid(0)
    print(x.result())

    for i in range(0, n):
        foo = random_uuid(0)
        print(foo.result())
        assert foo.result() == x.result(), "Memoized results were not used"



from parsl.config import Config
from parsl.executors.taskvine import TaskVineExecutor, TaskVineManagerConfig
from parsl.providers import LocalProvider

config = Config(
    executors=[TaskVineExecutor(label='VineExec',
                                worker_launch_method='factory',
                                manager_config=TaskVineManagerConfig(port=50055),
                                )]
)



from parsl.config import Config
from parsl.executors.threads import ThreadPoolExecutor


def fresh_config():
    tpe = ThreadPoolExecutor(label='local_threads_checkpoint_periodic', max_threads=1)
    return Config(
        executors=[tpe],
        checkpoint_mode='periodic',
        checkpoint_period='00:00:02'
    )



from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.launchers import SrunLauncher
from parsl.providers import SlurmProvider
from parsl.usage_tracking.levels import LEVEL_1

config = Config(
    executors=[
        HighThroughputExecutor(
            label='Anvil_GPU_Executor',
            working_dir='/anvil/scratch/path/to/your/directory',
            max_workers_per_node=1,
            provider=SlurmProvider(
                partition='gpu',
                account='YOUR_ALLOCATION_ON_ANVIL',
                nodes_per_block=1,
                cores_per_node=4,
                mem_per_node=50,
                init_blocks=1,
                max_blocks=1,
                walltime='00:30:00',
                # string to prepend to #SBATCH blocks in the submit
                # script to the scheduler
                scheduler_options='#SBATCH --gres gpu:1 --gpus-per-node=1',
                # Command to be run before starting a worker, such as:
                # 'module load Anaconda; source activate parsl_env'.
                worker_init='module load modtree/gpu',
                exclusive=False,
                launcher=SrunLauncher(),
            ),
        )
    ],
    usage_tracking=LEVEL_1,
)



from parsl.errors import ParslError


class DeserializationError(ParslError):
    """Failure at the deserialization of results/exceptions from remote workers.
    """

    def __init__(self, reason: str) -> None:
        self.reason = reason

    def __str__(self) -> str:
        return f"Failed to deserialize objects. Reason: {self.reason}"


class SerializationError(ParslError):
    """Failure to serialize task objects.
    """

    def __init__(self, fname: str) -> None:
        self.fname = fname
        self.troubleshooting = "https://parsl.readthedocs.io/en/latest/faq.html#addressing-serializationerror"

    def __str__(self) -> str:
        return f"Failed to serialize objects for an invocation of function {self.fname}. Refer {self.troubleshooting}"


class DeserializerPluginError(ParslError):
    """Failure to dynamically load a deserializer plugin.
    """

    def __init__(self, header: bytes) -> None:
        self.header = header

    def __str__(self) -> str:
        return f"Failed to load deserializer plugin for header {self.header!r}"



import contextlib
import os

import pytest

import parsl
from parsl import python_app
from parsl.tests.configs.local_threads_checkpoint import fresh_config


@contextlib.contextmanager
def parsl_configured(run_dir, **kw):
    c = fresh_config()
    c.run_dir = run_dir
    for config_attr, config_val in kw.items():
        setattr(c, config_attr, config_val)
    dfk = parsl.load(c)
    for ex in dfk.executors.values():
        ex.working_dir = run_dir
    yield dfk

    parsl.dfk().cleanup()


@python_app(cache=True)
def uuid_app():
    import uuid
    return uuid.uuid4()


@pytest.mark.local
def test_loading_checkpoint(tmpd_cwd):
    """Load memoization table from previous checkpoint
    """
    with parsl_configured(tmpd_cwd, checkpoint_mode="task_exit"):
        checkpoint_files = [os.path.join(parsl.dfk().run_dir, "checkpoint")]
        result = uuid_app().result()

    with parsl_configured(tmpd_cwd, checkpoint_files=checkpoint_files):
        relaunched = uuid_app().result()

    assert result == relaunched, "Expected following call to uuid_app to return cached uuid"



#!/usr/bin/env python3

import argparse
import logging
import math
import multiprocessing
import os
import pickle
import platform
import queue
import subprocess
import sys
import threading
import time
import uuid
from importlib.metadata import distributions
from multiprocessing.context import SpawnProcess
from multiprocessing.managers import DictProxy
from multiprocessing.sharedctypes import Synchronized
from typing import Dict, List, Optional, Sequence

import psutil
import zmq

from parsl import curvezmq
from parsl.addresses import tcp_url
from parsl.app.errors import RemoteExceptionWrapper
from parsl.executors.execute_task import execute_task
from parsl.executors.high_throughput.errors import WorkerLost
from parsl.executors.high_throughput.mpi_prefix_composer import (
    VALID_LAUNCHERS,
    compose_all,
)
from parsl.executors.high_throughput.mpi_resource_management import (
    MPITaskScheduler,
    TaskScheduler,
)
from parsl.executors.high_throughput.probe import probe_addresses
from parsl.multiprocessing import SpawnContext
from parsl.process_loggers import wrap_with_logs
from parsl.serialize import serialize
from parsl.version import VERSION as PARSL_VERSION

HEARTBEAT_CODE = (2 ** 32) - 1
DRAINED_CODE = (2 ** 32) - 2


class Manager:
    """ Manager manages task execution by the workers

                |         zmq              |    Manager         |   Worker Processes
                |                          |                    |
                | <-----Register with -----+                    |      Request task<--+
                |       N task capacity    |                    |          |          |
    Interchange | -------------------------+->Receive task batch|          |          |
                |                          |  Distribute tasks--+----> Get(block) &   |
                |                          |                    |      Execute task   |
                |                          |                    |          |          |
                | <------------------------+--Return results----+----  Post result    |
                |                          |                    |          |          |
                |                          |                    |          +----------+
                |                          |                IPC-Qeueues

    """
    def __init__(self, *,
                 addresses,
                 address_probe_timeout,
                 port,
                 cores_per_worker,
                 mem_per_worker,
                 max_workers_per_node,
                 prefetch_capacity,
                 uid,
                 block_id,
                 heartbeat_threshold,
                 heartbeat_period,
                 poll_period,
                 cpu_affinity,
                 enable_mpi_mode: bool = False,
                 mpi_launcher: str = "mpiexec",
                 available_accelerators: Sequence[str],
                 cert_dir: Optional[str],
                 drain_period: Optional[int]):
        """
        Parameters
        ----------
        addresses : str
             comma separated list of addresses for the interchange

        address_probe_timeout : int
             Timeout in seconds for the address probe to detect viable addresses
             to the interchange.

        uid : str
             string unique identifier

        block_id : str
             Block identifier that maps managers to the provider blocks they belong to.

        cores_per_worker : float
             cores to be assigned to each worker. Oversubscription is possible
             by setting cores_per_worker < 1.0.

        mem_per_worker : float
             GB of memory required per worker. If this option is specified, the node manager
             will check the available memory at startup and limit the number of workers such that
             the there's sufficient memory for each worker. If set to None, memory on node is not
             considered in the determination of workers to be launched on node by the manager.

        max_workers_per_node : int | float
             Caps the maximum number of workers that can be launched.

        prefetch_capacity : int
             Number of tasks that could be prefetched over available worker capacity.
             When there are a few tasks (<100) or when tasks are long running, this option should
             be set to 0 for better load balancing.

        heartbeat_threshold : int
             Seconds since the last message from the interchange after which the
             interchange is assumed to be un-available, and the manager initiates shutdown.

             Number of seconds since the last message from the interchange after which the worker
             assumes that the interchange is lost and the manager shuts down.

        heartbeat_period : int
             Number of seconds after which a heartbeat message is sent to the interchange, and workers
             are checked for liveness.

        poll_period : int
             Timeout period used by the manager in milliseconds.

        cpu_affinity : str
             Whether or how each worker should force its affinity to different CPUs

        available_accelerators: list of str
            List of accelerators available to the workers.

        enable_mpi_mode: bool
            When set to true, the manager assumes ownership of the batch job and each worker
            claims a subset of nodes from a shared pool to execute multi-node mpi tasks. Node
            info is made available to workers via env vars.

        mpi_launcher: str
            Set to one of the supported MPI launchers: ("srun", "aprun", "mpiexec")

        cert_dir : str | None
            Path to the certificate directory.

        drain_period: int | None
            Number of seconds to drain after  TODO: could be a nicer timespec involving m,s,h qualifiers for user friendliness?
        """

        logger.info("Manager initializing")

        self._start_time = time.time()

        self.cert_dir = cert_dir
        self.zmq_context = curvezmq.ClientContext(self.cert_dir)

        addresses = ','.join(tcp_url(a, port) for a in addresses.split(','))
        try:
            self._ix_url = probe_addresses(
                self.zmq_context,
                addresses,
                timeout_ms=1_000 * address_probe_timeout,
                identity=uid.encode('utf-8'),
            )
        except ConnectionError:
            addys = ", ".join(addresses.split(","))
            logger.error(f"Unable to connect to interchange; attempted addresses: {addys}")
            raise

        logger.info(f"Probe discovered interchange url: {self._ix_url}")

        self.uid = uid
        self.block_id = block_id
        self.start_time = time.time()

        self.enable_mpi_mode = enable_mpi_mode
        self.mpi_launcher = mpi_launcher

        if os.environ.get('PARSL_CORES'):
            cores_on_node = int(os.environ['PARSL_CORES'])
        else:
            cores_on_node = SpawnContext.cpu_count()

        if os.environ.get('PARSL_MEMORY_GB'):
            available_mem_on_node = float(os.environ['PARSL_MEMORY_GB'])
        else:
            available_mem_on_node = round(psutil.virtual_memory().available / (2**30), 1)

        self.max_workers_per_node = max_workers_per_node
        self.prefetch_capacity = prefetch_capacity

        mem_slots = max_workers_per_node
        # Avoid a divide by 0 error.
        if mem_per_worker and mem_per_worker > 0:
            mem_slots = math.floor(available_mem_on_node / mem_per_worker)

        self.worker_count: int = min(max_workers_per_node,
                                     mem_slots,
                                     math.floor(cores_on_node / cores_per_worker))

        self._mp_manager = SpawnContext.Manager()  # Starts a server process
        self._tasks_in_progress = self._mp_manager.dict()
        self._stop_event = threading.Event()  # when set, will begin shutdown process

        self.monitoring_queue = self._mp_manager.Queue()
        self.pending_task_queue = SpawnContext.Queue()
        self.pending_result_queue = SpawnContext.Queue()
        self.task_scheduler: TaskScheduler
        if self.enable_mpi_mode:
            self.task_scheduler = MPITaskScheduler(
                self.pending_task_queue,
                self.pending_result_queue,
            )
        else:
            self.task_scheduler = TaskScheduler(
                self.pending_task_queue,
                self.pending_result_queue
            )
        self.ready_worker_count = SpawnContext.Value("i", 0)

        self.max_queue_size = self.prefetch_capacity + self.worker_count

        self.tasks_per_round = 1

        self.heartbeat_period = heartbeat_period
        self.heartbeat_threshold = heartbeat_threshold
        self.poll_period = poll_period

        self.drain_time: float
        if drain_period:
            self.drain_time = self._start_time + drain_period
            logger.info(f"Will request drain at {self.drain_time}")
        else:
            self.drain_time = float('inf')

        self.cpu_affinity = cpu_affinity

        # Define accelerator available, adjust worker count accordingly
        self.available_accelerators = available_accelerators
        self.accelerators_available = len(available_accelerators) > 0
        if self.accelerators_available:
            self.worker_count = min(len(self.available_accelerators), self.worker_count)
        logger.info("Manager will spawn {} workers".format(self.worker_count))

    def create_reg_message(self) -> dict:
        """ Creates a registration message to identify the worker to the interchange
        """
        return {
            'type': 'registration',
            'parsl_v': PARSL_VERSION,
            'python_v': "{}.{}.{}".format(
                sys.version_info.major,
                sys.version_info.minor,
                sys.version_info.micro
            ),
            'packages': {d.metadata['Name']: d.version for d in distributions()},
            'worker_count': self.worker_count,
            'uid': self.uid,
            'block_id': self.block_id,
            'start_time': self.start_time,
            'prefetch_capacity': self.prefetch_capacity,
            'max_capacity': self.worker_count + self.prefetch_capacity,
            'os': platform.system(),
            'hostname': platform.node(),
            'dir': os.getcwd(),
            'cpu_count': psutil.cpu_count(logical=False),
            'total_memory': psutil.virtual_memory().total,
        }

    @staticmethod
    def heartbeat_to_incoming(task_incoming: zmq.Socket) -> None:
        """ Send heartbeat to the incoming task queue
        """
        # don't need to dumps and encode this every time - could do as a global on import?
        b_msg = pickle.dumps({'type': 'heartbeat'})
        task_incoming.send(b_msg)
        logger.debug("Sent heartbeat")

    @staticmethod
    def drain_to_incoming(task_incoming: zmq.Socket) -> None:
        """ Send heartbeat to the incoming task queue
        """
        msg = {'type': 'drain'}
        b_msg = pickle.dumps(msg)
        task_incoming.send(b_msg)
        logger.debug("Sent drain")

    @wrap_with_logs
    def interchange_communicator(self, pair_setup: threading.Event):
        """ Pull tasks from the incoming tasks zmq pipe onto the internal
        pending task queue
        """
        logger.info("starting")

        results_sock = self.zmq_context.socket(zmq.PAIR)
        results_sock.setsockopt(zmq.LINGER, 0)
        results_sock.bind("inproc://results")
        pair_setup.set()

        # Linger is set to 0, so that the manager can exit even when there might be
        # messages in the pipe
        ix_sock = self.zmq_context.socket(zmq.DEALER)
        ix_sock.setsockopt(zmq.IDENTITY, self.uid.encode('utf-8'))
        ix_sock.setsockopt(zmq.LINGER, 0)
        ix_sock.connect(self._ix_url)
        logger.info("Manager task pipe connected to interchange")

        poller = zmq.Poller()
        poller.register(results_sock, zmq.POLLIN)
        poller.register(ix_sock, zmq.POLLIN)

        ix_sock.send(pickle.dumps({"type": "connection_probe"}))
        evts = dict(poller.poll(timeout=self.heartbeat_period))
        if evts.get(ix_sock) is None:
            logger.error(f"Failed to connect to interchange ({self._ix_url}")

        ix_sock.recv()
        logger.info(f"Successfully connected to interchange via URL: {self._ix_url}")

        # Send a registration message
        msg = self.create_reg_message()
        logger.debug("Sending registration message: %s", msg)
        ix_sock.send(pickle.dumps(msg))
        last_beat = time.time()
        last_interchange_contact = time.time()
        task_recv_counter = 0

        while not self._stop_event.is_set():

            # This loop will sit inside poller.poll until either a message
            # arrives or one of these event times is reached. This code
            # assumes that the event times won't change except on iteration
            # of this loop - so will break if a different thread does
            # anything to bring one of the event times earlier - and that the
            # time here are correctly copy-pasted from the relevant if
            # statements.
            next_interesting_event_time = min(last_beat + self.heartbeat_period,
                                              self.drain_time,
                                              last_interchange_contact + self.heartbeat_threshold)
            try:
                pending_task_count = self.pending_task_queue.qsize()
            except NotImplementedError:
                # Ref: https://github.com/python/cpython/blob/6d5e0dc0e330f4009e8dc3d1642e46b129788877/Lib/multiprocessing/queues.py#L125
                pending_task_count = f"pending task count is not available on {platform.system()}"  # type: ignore[assignment]

            logger.debug(
                'ready workers: %d, pending tasks: %d',
                self.ready_worker_count.value,  # type: ignore[attr-defined]
                pending_task_count,
            )

            if time.time() >= last_beat + self.heartbeat_period:
                self.heartbeat_to_incoming(ix_sock)
                last_beat = time.time()

            if time.time() > self.drain_time:
                logger.info("Requesting drain")
                self.drain_to_incoming(ix_sock)
                # This will start the pool draining...
                # Drained exit behaviour does not happen here. It will be
                # driven by the interchange sending a DRAINED_CODE message.

                # now set drain time to the far future so we don't send a drain
                # message every iteration.
                self.drain_time = float('inf')

            poll_duration_s = max(0, next_interesting_event_time - time.time())
            socks = dict(poller.poll(timeout=poll_duration_s * 1000))

            if socks.get(ix_sock) == zmq.POLLIN:
                pkl_msg = ix_sock.recv()
                tasks = pickle.loads(pkl_msg)
                last_interchange_contact = time.time()

                if tasks == HEARTBEAT_CODE:
                    logger.debug("Got heartbeat from interchange")
                elif tasks == DRAINED_CODE:
                    logger.info("Got fully drained message from interchange - setting kill flag")
                    self._stop_event.set()
                else:
                    task_recv_counter += len(tasks)
                    logger.debug("Got executor tasks: {}, cumulative count of tasks: {}".format(
                        [t['task_id'] for t in tasks], task_recv_counter
                    ))

                    for task in tasks:
                        self.task_scheduler.put_task(task)

            elif socks.get(results_sock) == zmq.POLLIN:
                meta_b = pickle.dumps({'type': 'result'})
                ix_sock.send_multipart([meta_b, results_sock.recv()])
                logger.debug("Result sent to interchange")

            else:
                logger.debug("No incoming tasks")

                # Only check if no messages were received.
                if time.time() >= last_interchange_contact + self.heartbeat_threshold:
                    logger.critical("Missing contact with interchange beyond heartbeat_threshold")
                    self._stop_event.set()
                    logger.critical("Exiting")
                    break

        ix_sock.close()
        logger.info("Exiting")

    @wrap_with_logs
    def ferry_result(self, may_connect: threading.Event):
        """ Listens on the pending_result_queue and ferries results to the interchange
         connected thread
        """
        logger.debug("Begin")

        # Linger is set to 0, so that the manager can exit even when there might be
        # messages in the pipe
        notify_sock = self.zmq_context.socket(zmq.PAIR)
        notify_sock.setsockopt(zmq.LINGER, 0)
        may_connect.wait()
        notify_sock.connect("inproc://results")

        while not self._stop_event.is_set():
            try:
                r = self.task_scheduler.get_result()
                if r is None:
                    continue
                logger.debug("Result received from worker")
                notify_sock.send(r)
            except Exception:
                logger.exception("Failed to send result to interchange")

        notify_sock.close()
        logger.debug("Exiting")

    def worker_watchdog(self, procs: dict[int, SpawnProcess]):
        """Keeps workers alive."""
        logger.debug("Starting worker watchdog")

        while not self._stop_event.wait(self.heartbeat_period):
            for worker_id, p in procs.items():
                if not p.is_alive():
                    logger.error("Worker {} has died".format(worker_id))
                    try:
                        task = self._tasks_in_progress.pop(worker_id)
                        logger.info("Worker {} was busy when it died".format(worker_id))
                        try:
                            raise WorkerLost(worker_id, platform.node())
                        except Exception:
                            logger.info("Putting exception for executor task {} in the pending result queue".format(task['task_id']))
                            result_package = {'type': 'result',
                                              'task_id': task['task_id'],
                                              'exception': serialize(RemoteExceptionWrapper(*sys.exc_info()))}
                            pkl_package = pickle.dumps(result_package)
                            self.pending_result_queue.put(pkl_package)
                    except KeyError:
                        logger.info("Worker {} was not busy when it died".format(worker_id))

                    procs[worker_id] = self._start_worker(worker_id)
                    logger.info("Worker {} has been restarted".format(worker_id))

        logger.debug("Exiting")

    @wrap_with_logs
    def handle_monitoring_messages(self):
        """Transfer messages from the managed monitoring queue to the result queue.

        We separate the queues so that the result queue does not rely on a manager
        process, which adds overhead that causes slower queue operations but enables
        use across processes started in fork and spawn contexts.

        We transfer the messages to the result queue to reuse the ZMQ connection between
        the manager and the interchange.
        """
        logger.debug("Starting monitoring handler thread")

        while not self._stop_event.is_set():
            try:
                logger.debug("Starting monitor_queue.get()")
                msg = self.monitoring_queue.get(block=True)
                if msg is None:
                    continue
                logger.debug("Got a monitoring message")
                self.pending_result_queue.put(msg)
                logger.debug("Put monitoring message on pending_result_queue")
            except Exception:
                logger.exception("Failed to forward monitoring message")

        logger.debug("Exiting")

    def start(self):
        """ Start the worker processes.

        TODO: Move task receiving to a thread
        """
        procs: dict[int, SpawnProcess] = {}
        for worker_id in range(self.worker_count):
            procs[worker_id] = self._start_worker(worker_id)

        logger.debug("Workers started")

        pair_setup = threading.Event()

        thr_task_puller = threading.Thread(
            target=self.interchange_communicator,
            args=(pair_setup,),
            name="Interchange-Communicator",
        )
        thr_result_ferry = threading.Thread(
            target=self.ferry_result, args=(pair_setup,), name="Result-Shovel")
        thr_worker_watchdog = threading.Thread(
            target=self.worker_watchdog, args=(procs,), name="worker-watchdog"
        )
        thr_monitoring_handler = threading.Thread(
            target=self.handle_monitoring_messages, name="Monitoring-Handler"
        )

        thr_task_puller.start()
        thr_result_ferry.start()
        thr_worker_watchdog.start()
        thr_monitoring_handler.start()

        logger.info("Manager threads started")

        # This might need a multiprocessing event to signal back.
        self._stop_event.wait()
        logger.info("Stop event set; terminating worker processes")

        # Invite blocking threads to quit
        self.monitoring_queue.put(None)
        self.pending_result_queue.put(None)

        thr_monitoring_handler.join()
        thr_worker_watchdog.join()
        thr_result_ferry.join()
        thr_task_puller.join()

        for worker_id in procs:
            p = procs[worker_id]
            proc_info = f"(PID: {p.pid}, Worker ID: {worker_id})"
            logger.debug(f"Signaling worker {p.name} (TERM). {proc_info}")
            p.terminate()

        self.zmq_context.term()

        # give processes 1 second to gracefully shut themselves down, based on the
        # SIGTERM (.terminate()) just sent; after then, we pull the plug.
        force_child_shutdown_at = time.monotonic() + 1
        while procs:
            worker_id, p = procs.popitem()
            timeout = max(force_child_shutdown_at - time.monotonic(), 0.000001)
            p.join(timeout=timeout)
            proc_info = f"(PID: {p.pid}, Worker ID: {worker_id})"
            if p.exitcode is not None:
                logger.debug(
                    "Worker joined successfully.  %s (exitcode: %s)", proc_info, p.exitcode
                )

            else:
                logger.warning(
                    f"Worker {p.name} ({worker_id}) failed to terminate in a timely"
                    f" manner; sending KILL signal to process. {proc_info}"
                )
                p.kill()
                p.join()
            p.close()

        delta = time.time() - self._start_time
        logger.info("process_worker_pool ran for {} seconds".format(delta))

    def _start_worker(self, worker_id: int) -> SpawnProcess:
        p = SpawnContext.Process(
            target=worker,
            args=(
                worker_id,
                self.uid,
                self.worker_count,
                self.pending_task_queue,
                self.pending_result_queue,
                self.monitoring_queue,
                self.ready_worker_count,
                self._tasks_in_progress,
                self.cpu_affinity,
                self.available_accelerators[worker_id] if self.accelerators_available else None,
                self.block_id,
                self.heartbeat_period,
                os.getpid(),
                args.logdir,
                args.debug,
                self.mpi_launcher,
            ),
            name="HTEX-Worker-{}".format(worker_id),
        )
        p.start()
        return p


def update_resource_spec_env_vars(mpi_launcher: str, resource_spec: Dict, node_info: List[str]) -> None:
    prefix_table = compose_all(mpi_launcher, resource_spec=resource_spec, node_hostnames=node_info)
    for key in prefix_table:
        os.environ[key] = prefix_table[key]


def _init_mpi_env(mpi_launcher: str, resource_spec: Dict):
    node_list = resource_spec.get("MPI_NODELIST")
    if node_list is None:
        return
    nodes_for_task = node_list.split(',')
    logger.info(f"Launching task on provisioned nodes: {nodes_for_task}")
    update_resource_spec_env_vars(mpi_launcher=mpi_launcher, resource_spec=resource_spec, node_info=nodes_for_task)


@wrap_with_logs(target="worker_log")
def worker(
    worker_id: int,
    pool_id: str,
    pool_size: int,
    task_queue: multiprocessing.Queue,
    result_queue: multiprocessing.Queue,
    monitoring_queue: queue.Queue,
    ready_worker_count: Synchronized,
    tasks_in_progress: DictProxy,
    cpu_affinity: str,
    accelerator: Optional[str],
    block_id: str,
    task_queue_timeout: int,
    manager_pid: int,
    logdir: str,
    debug: bool,
    mpi_launcher: str,
):
    # override the global logger inherited from the __main__ process (which
    # usually logs to manager.log) with one specific to this worker.
    global logger
    logger = start_file_logger('{}/block-{}/{}/worker_{}.log'.format(logdir, block_id, pool_id, worker_id),
                               worker_id,
                               name="worker_log",
                               level=logging.DEBUG if debug else logging.INFO)

    # Store worker ID as an environment variable
    os.environ['PARSL_WORKER_RANK'] = str(worker_id)
    os.environ['PARSL_WORKER_COUNT'] = str(pool_size)
    os.environ['PARSL_WORKER_POOL_ID'] = str(pool_id)
    os.environ['PARSL_WORKER_BLOCK_ID'] = str(block_id)

    import parsl.executors.high_throughput.monitoring_info as mi
    mi.result_queue = monitoring_queue

    logger.info('Worker {} started'.format(worker_id))
    if debug:
        logger.debug("Debug logging enabled")

    # If desired, set process affinity
    if cpu_affinity != "none":
        # Count the number of cores per worker
        # OSX does not implement os.sched_getaffinity
        avail_cores = sorted(os.sched_getaffinity(0))  # type: ignore[attr-defined, unused-ignore]
        cores_per_worker = len(avail_cores) // pool_size
        assert cores_per_worker > 0, "Affinity does not work if there are more workers than cores"

        # Determine this worker's cores
        if cpu_affinity == "block":
            my_cores = avail_cores[cores_per_worker * worker_id:cores_per_worker * (worker_id + 1)]
        elif cpu_affinity == "block-reverse":
            cpu_worker_id = pool_size - worker_id - 1  # To assign in reverse order
            my_cores = avail_cores[cores_per_worker * cpu_worker_id:cores_per_worker * (cpu_worker_id + 1)]
        elif cpu_affinity == "alternating":
            my_cores = avail_cores[worker_id::pool_size]
        elif cpu_affinity[0:4] == "list":
            thread_ranks = cpu_affinity.split(":")[1:]
            if len(thread_ranks) != pool_size:
                raise ValueError("Affinity list {} has wrong number of thread ranks".format(cpu_affinity))
            threads = thread_ranks[worker_id]
            thread_list = threads.split(",")
            my_cores = []
            for tl in thread_list:
                thread_range = tl.split("-")
                if len(thread_range) == 1:
                    my_cores.append(int(thread_range[0]))
                elif len(thread_range) == 2:
                    start_thread = int(thread_range[0])
                    end_thread = int(thread_range[1]) + 1
                    my_cores += list(range(start_thread, end_thread))
                else:
                    raise ValueError("Affinity list formatting is not expected {}".format(cpu_affinity))
        else:
            raise ValueError("Affinity strategy {} is not supported".format(cpu_affinity))

        # Set the affinity for OpenMP
        #  See: https://hpc-tutorials.llnl.gov/openmp/ProcessThreadAffinity.pdf
        proc_list = ",".join(map(str, my_cores))
        os.environ["OMP_NUM_THREADS"] = str(len(my_cores))
        os.environ["GOMP_CPU_AFFINITY"] = proc_list  # Compatible with GCC OpenMP
        os.environ["KMP_AFFINITY"] = f"explicit,proclist=[{proc_list}]"  # For Intel OpenMP

        # Set the affinity for this worker
        # OSX does not implement os.sched_setaffinity so type checking
        # is ignored here in two ways:
        # On a platform without sched_setaffinity, that attribute will not
        # be defined, so ignore[attr-defined] will tell mypy to ignore this
        # incorrect-for-OS X attribute access.
        # On a platform with sched_setaffinity, that type: ignore message
        # will be redundant, and ignore[unused-ignore] tells mypy to ignore
        # that this ignore is unneeded.
        os.sched_setaffinity(0, my_cores)  # type: ignore[attr-defined, unused-ignore]
        logger.info("Set worker CPU affinity to {}".format(my_cores))

    # If desired, pin to accelerator
    if accelerator is not None:

        # If CUDA devices, find total number of devices to allow for MPS
        # See: https://developer.nvidia.com/system-management-interface
        nvidia_smi_cmd = "nvidia-smi -L > /dev/null && nvidia-smi -L | wc -l"
        nvidia_smi_ret = subprocess.run(nvidia_smi_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if nvidia_smi_ret.returncode == 0:
            num_cuda_devices = int(nvidia_smi_ret.stdout.split()[0])
        else:
            num_cuda_devices = None

        try:
            if num_cuda_devices is not None:
                procs_per_cuda_device = pool_size // num_cuda_devices
                partitioned_accelerator = str(int(accelerator) // procs_per_cuda_device)  # multiple workers will share a GPU
                os.environ["CUDA_VISIBLE_DEVICES"] = partitioned_accelerator
                logger.info(f'Pinned worker to partitioned cuda device: {partitioned_accelerator}')
            else:
                os.environ["CUDA_VISIBLE_DEVICES"] = accelerator
        except (TypeError, ValueError, ZeroDivisionError):
            os.environ["CUDA_VISIBLE_DEVICES"] = accelerator
        os.environ["ROCR_VISIBLE_DEVICES"] = accelerator
        os.environ["ZE_AFFINITY_MASK"] = accelerator
        os.environ["ZE_ENABLE_PCI_ID_DEVICE_ORDER"] = '1'

        logger.info(f'Pinned worker to accelerator: {accelerator}')

    def manager_is_alive():
        try:
            # This does not kill the process, but instead raises
            # an exception if the process doesn't exist
            os.kill(manager_pid, 0)
        except OSError:
            logger.critical(f"Manager ({manager_pid}) died; worker {worker_id} shutting down")
            return False
        else:
            return True

    worker_enqueued = False
    while manager_is_alive():
        if not worker_enqueued:
            with ready_worker_count.get_lock():
                ready_worker_count.value += 1
            worker_enqueued = True

        try:
            # The worker will receive {'task_id':<tid>, 'buffer':<buf>}
            req = task_queue.get(timeout=task_queue_timeout)
        except queue.Empty:
            continue

        tasks_in_progress[worker_id] = req
        tid = req['task_id']
        logger.info("Received executor task {}".format(tid))

        with ready_worker_count.get_lock():
            ready_worker_count.value -= 1
        worker_enqueued = False

        _init_mpi_env(mpi_launcher=mpi_launcher, resource_spec=req["resource_spec"])

        try:
            result = execute_task(req['buffer'])
            serialized_result = serialize(result, buffer_threshold=1000000)
        except Exception as e:
            logger.info('Caught an exception: {}'.format(e))
            result_package = {'type': 'result', 'task_id': tid, 'exception': serialize(RemoteExceptionWrapper(*sys.exc_info()))}
        else:
            result_package = {'type': 'result', 'task_id': tid, 'result': serialized_result}
            # logger.debug("Result: {}".format(result))

        logger.info("Completed executor task {}".format(tid))
        try:
            pkl_package = pickle.dumps(result_package)
        except Exception:
            logger.exception("Caught exception while trying to pickle the result package")
            pkl_package = pickle.dumps({'type': 'result', 'task_id': tid,
                                        'exception': serialize(RemoteExceptionWrapper(*sys.exc_info()))
                                        })

        result_queue.put(pkl_package)
        tasks_in_progress.pop(worker_id)
        logger.info("All processing finished for executor task {}".format(tid))


def start_file_logger(filename, rank, name='parsl', level=logging.DEBUG, format_string=None):
    """Add a stream log handler.

    Args:
        - filename (string): Name of the file to write logs to
        - name (string): Logger name
        - level (logging.LEVEL): Set the logging level.
        - format_string (string): Set the format string

    Returns:
       -  None
    """
    if format_string is None:
        format_string = "%(asctime)s.%(msecs)03d %(name)s:%(lineno)d " \
                        "%(process)d %(threadName)s " \
                        "[%(levelname)s]  %(message)s"

    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    handler = logging.FileHandler(filename)
    handler.setLevel(level)
    formatter = logging.Formatter(format_string, datefmt='%Y-%m-%d %H:%M:%S')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def get_arg_parser() -> argparse.ArgumentParser:

    def strategyorlist(s: str):
        s = s.lower()
        allowed_strategies = ("none", "block", "alternating", "block-reverse")
        if s in allowed_strategies:
            return s
        elif s[0:4] == "list":
            return s
        err_msg = f"cpu-affinity must be one of {allowed_strategies} or a list format"
        raise argparse.ArgumentTypeError(err_msg)

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d", "--debug", action='store_true', help="Enable logging at DEBUG level",
    )
    parser.add_argument(
        "-a",
        "--addresses",
        required=True,
        help="Comma separated list of addresses at which the interchange could be reached",
    )
    parser.add_argument(
        "--cert_dir", required=True, help="Path to certificate directory."
    )
    parser.add_argument(
        "-l",
        "--logdir",
        default="process_worker_pool_logs",
        help="Process worker pool log directory",
    )
    parser.add_argument(
        "-u",
        "--uid",
        default=str(uuid.uuid4()).split('-')[-1],
        help="Unique identifier string for Manager",
    )
    parser.add_argument(
        "-b", "--block_id", default=None, help="Block identifier for Manager"
    )
    parser.add_argument(
        "-c",
        "--cores_per_worker",
        default="1.0",
        help="Number of cores assigned to each worker process. Default=1.0",
    )
    parser.add_argument(
        "-m",
        "--mem_per_worker",
        default=0,
        help="GB of memory assigned to each worker process. Default=0, no assignment",
    )
    parser.add_argument(
        "-P",
        "--port",
        required=True,
        help="Port for communication with the interchange",
    )
    parser.add_argument(
        "--max_workers_per_node",
        default=float('inf'),
        help="Caps the maximum workers that can be launched, default:infinity",
    )
    parser.add_argument(
        "-p",
        "--prefetch_capacity",
        default=0,
        help="Number of tasks that can be prefetched to the manager. Default is 0.",
    )
    parser.add_argument(
        "--hb_period",
        default=30,
        help="Heartbeat period in seconds. Uses manager default unless set",
    )
    parser.add_argument(
        "--hb_threshold",
        default=120,
        help="Heartbeat threshold in seconds. Uses manager default unless set",
    )
    parser.add_argument(
        "--drain_period",
        default=None,
        help="Drain this pool after specified number of seconds. By default, does not drain.",
    )
    parser.add_argument(
        "--address_probe_timeout",
        default=30,
        help="Timeout to probe for viable address to interchange. Default: 30s",
    )
    parser.add_argument(
        "--poll", default=10, help="Poll period used in milliseconds"
    )
    parser.add_argument(
        "--cpu-affinity",
        type=strategyorlist,
        required=True,
        help="Whether/how workers should control CPU affinity.",
    )
    parser.add_argument(
        "--available-accelerators",
        type=str,
        nargs="*",
        default=[],
        help="Names of available accelerators, if not given assumed to be zero accelerators available",
    )
    parser.add_argument(
        "--enable_mpi_mode", action='store_true', help="Enable MPI mode"
    )
    parser.add_argument(
        "--mpi-launcher",
        type=str,
        choices=VALID_LAUNCHERS,
        help="MPI launcher to use iff enable_mpi_mode=true",
    )

    return parser


if __name__ == "__main__":
    parser = get_arg_parser()
    args = parser.parse_args()

    os.makedirs(os.path.join(args.logdir, "block-{}".format(args.block_id), args.uid), exist_ok=True)

    logger = start_file_logger(
        f'{args.logdir}/block-{args.block_id}/{args.uid}/manager.log',
        0,
        level=logging.DEBUG if args.debug is True else logging.INFO
    )
    logger.info(
        f"\n  Python version: {sys.version}"
        f"\n  Debug logging: {args.debug}"
        f"\n  Certificates dir: {args.cert_dir}"
        f"\n  Log dir: {args.logdir}"
        f"\n  Manager ID: {args.uid}"
        f"\n  Block ID: {args.block_id}"
        f"\n  cores_per_worker: {args.cores_per_worker}"
        f"\n  mem_per_worker: {args.mem_per_worker}"
        f"\n  Interchange port: {args.port}"
        f"\n  addresses: {args.addresses}"
        f"\n  max_workers_per_node: {args.max_workers_per_node}"
        f"\n  poll_period: {args.poll}"
        f"\n  address_probe_timeout: {args.address_probe_timeout}"
        f"\n  Prefetch capacity: {args.prefetch_capacity}"
        f"\n  Heartbeat threshold: {args.hb_threshold}"
        f"\n  Heartbeat period: {args.hb_period}"
        f"\n  Drain period: {args.drain_period}"
        f"\n  CPU affinity: {args.cpu_affinity}"
        f"\n  Accelerators: {' '.join(args.available_accelerators)}"
        f"\n  enable_mpi_mode: {args.enable_mpi_mode}"
        f"\n  mpi_launcher: {args.mpi_launcher}"
    )
    try:
        manager = Manager(port=args.port,
                          addresses=args.addresses,
                          address_probe_timeout=int(args.address_probe_timeout),
                          uid=args.uid,
                          block_id=args.block_id,
                          cores_per_worker=float(args.cores_per_worker),
                          mem_per_worker=None if args.mem_per_worker == 'None' else float(args.mem_per_worker),
                          max_workers_per_node=(
                              args.max_workers_per_node if args.max_workers_per_node == float('inf')
                              else int(args.max_workers_per_node)
                          ),
                          prefetch_capacity=int(args.prefetch_capacity),
                          heartbeat_threshold=int(args.hb_threshold),
                          heartbeat_period=int(args.hb_period),
                          drain_period=None if args.drain_period == "None" else int(args.drain_period),
                          poll_period=int(args.poll),
                          cpu_affinity=args.cpu_affinity,
                          enable_mpi_mode=args.enable_mpi_mode,
                          mpi_launcher=args.mpi_launcher,
                          available_accelerators=args.available_accelerators,
                          cert_dir=None if args.cert_dir == "None" else args.cert_dir)
        manager.start()

    except Exception:
        logger.critical("Process worker pool exiting with an exception", exc_info=True)
        raise
    else:
        logger.info("Process worker pool exiting normally")
        print("Process worker pool exiting normally")
